#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""gflownet trainer unit tests."""

import os
import shutil
import tempfile
from typing import Any, Dict, cast

import numpy as np
import pytest

from gt4sd.frameworks.gflownet.envs.graph_building_env import GraphBuildingEnv
from gt4sd.frameworks.gflownet.envs.mol_building_env import MolBuildingEnvContext
from gt4sd.frameworks.gflownet.tests.qm9 import QM9Dataset, QM9GapTask
from gt4sd.training_pipelines import TRAINING_PIPELINE_MAPPING, GFlowNetTrainingPipeline


def _create_training_output_filepaths(directory: str) -> Dict[str, str]:
    """Create output filepath from directory.

    Args:
        directory: output directory.

    Returns:
        a dictionary containing the output files.
    """
    return {
        "log_file": os.path.join(directory, "log.txt"),
        "model_save": os.path.join(directory, "model.pt"),
        "config_save": os.path.join(directory, "config.pt"),
    }


template_config = {
    "model_args": {
        "algorithm": "trajectory_balance",
        "context": None,
        "environment": None,
        "model": "graph_transformer_gfn",
        "sampling_model": "graph_transformer_gfn",
        "task": "qm9",
        "bootstrap_own_reward": False,
        "num_emb": 128,
        "num_layers": 4,
        "tb_epsilon": 1e-10,
        "illegal_action_logreward": -50,
        "reward_loss_multiplier": 1.0,
        "temperature_sample_dist": "uniform",
        "temperature_dist_params": "(.5, 32)",
        "weight_decay": 1e-8,
        "momentum": 0.9,
        "adam_eps": 1e-8,
        "lr_decay": 20000,
        "z_lr_decay": 20000,
        "clip_grad_type": "norm",
        "clip_grad_param": 10.0,
        "random_action_prob": 0.001,
        "sampling_tau": 0.0,
        "max_nodes": 9,
        "num_offline": 10,
    },
    "pl_trainer_args": {
        "strategy": "ddp",
        "basename": "gflownet",
        "check_val_every_n_epoch": 5,
        "trainer_log_every_n_steps": 50,
        "auto_lr_find": True,
        "profiler": "simple",
        "learning_rate": 0.0001,
        "test_output_path": "./test",
        "num_workers": 0,
        "log_dir": "./log/",
        "save_dir": "./log/",
        "num_training_steps": 1000,
        "validate_every": 1000,
        "seed": 142857,
        "device": "cpu",
        "development_mode": True,
        "resume_from_checkpoint": None,
        "save_top_k": 1,
        "epochs": 3,
    },
    "dataset_args": {
        "dataset": "qm9",
        "dataset_path": "./data/qm9.h5",
        "epoch": 100,
        "batch_size": 64,
        "global_batch_size": 16,
        "validation_split": None,
        "sampling_iterator": True,
        "ratio": 0.9,
    },
}


@pytest.mark.skip(reason="we need to add support for dataset buckets")
def test_train():

    pipeline = TRAINING_PIPELINE_MAPPING.get("gflownet-trainer")

    assert pipeline is not None

    TEMPORARY_DIRECTORY = tempfile.mkdtemp()

    test_pipeline = cast(GFlowNetTrainingPipeline, pipeline())

    config: Dict[str, Any] = template_config.copy()
    for key, value in _create_training_output_filepaths(TEMPORARY_DIRECTORY).items():
        config["pl_trainer_args"][key] = value
    config["pl_trainer_args"]["rng"] = np.random.default_rng(
        config["pl_trainer_args"]["seed"]
    )

    dataset = QM9Dataset(config["dataset_args"]["dataset_path"], target="gap")
    environment = GraphBuildingEnv()
    context = MolBuildingEnvContext()

    task = QM9GapTask(
        configuration={
            **config["pl_trainer_args"],
            **config["model_args"],
            **config["dataset_args"],
        },
        dataset=dataset,
    )

    config["dataset"] = dataset
    config["environment"] = environment
    config["context"] = context
    config["task"] = task

    test_pipeline.train(**config)

    shutil.rmtree(TEMPORARY_DIRECTORY)



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""TorchDrug tests."""

from typing import ClassVar, Type

import pytest

from gt4sd.algorithms.core import AlgorithmConfiguration
from gt4sd.algorithms.generation.torchdrug import (
    TorchDrugGCPN,
    TorchDrugGenerator,
    TorchDrugGraphAF,
)
from gt4sd.algorithms.registry import ApplicationsRegistry
from gt4sd.tests.utils import GT4SDTestSettings

test_settings = GT4SDTestSettings.get_instance()


def get_classvar_type(class_var):
    """Extract type from ClassVar type annotation: `ClassVar[T]] -> T`."""
    return class_var.__args__[0]


@pytest.mark.parametrize(
    "config_class, algorithm_type, domain, algorithm_name",
    [
        (
            TorchDrugGCPN,
            "generation",
            "materials",
            TorchDrugGenerator.__name__,
        ),
        (
            TorchDrugGraphAF,
            "generation",
            "materials",
            TorchDrugGenerator.__name__,
        ),
    ],
)
def test_config_class(
    config_class: Type[AlgorithmConfiguration],
    algorithm_type: str,
    domain: str,
    algorithm_name: str,
):
    assert config_class.algorithm_type == algorithm_type
    assert config_class.domain == domain
    assert config_class.algorithm_name == algorithm_name

    for keyword, type_annotation in config_class.__annotations__.items():
        if keyword in ("algorithm_type", "domain", "algorithm_name"):
            assert type_annotation.__origin__ is ClassVar  # type: ignore
            assert str == get_classvar_type(type_annotation)


@pytest.mark.parametrize(
    "config_class",
    [(TorchDrugGCPN), (TorchDrugGraphAF)],
)
def test_config_instance(config_class: Type[AlgorithmConfiguration]):
    config = config_class()  # type:ignore
    assert config.algorithm_application == config_class.__name__


@pytest.mark.parametrize(
    "config_class",
    [(TorchDrugGCPN), (TorchDrugGraphAF)],
)
def test_available_versions(config_class: Type[AlgorithmConfiguration]):
    versions = config_class.list_versions()
    assert len(versions) > 0


@pytest.mark.parametrize(
    "config, algorithm, version",
    [
        pytest.param(
            TorchDrugGCPN,
            TorchDrugGenerator,
            "zinc250k_v0",
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            TorchDrugGraphAF,
            TorchDrugGenerator,
            "zinc250k_v0",
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            TorchDrugGCPN,
            TorchDrugGenerator,
            "qed_v0",
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            TorchDrugGraphAF,
            TorchDrugGenerator,
            "qed_v0",
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            TorchDrugGCPN,
            TorchDrugGenerator,
            "plogp_v0",
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            TorchDrugGraphAF,
            TorchDrugGenerator,
            "plogp_v0",
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
    ],
)
def test_generation_via_import(config, algorithm, version: str):
    algorithm = algorithm(configuration=config(algorithm_version=version))
    items = list(algorithm.sample(1))
    assert len(items) == 1


@pytest.mark.parametrize(
    "algorithm_application, algorithm_type, domain, algorithm_version, algorithm_name",
    [
        pytest.param(
            TorchDrugGCPN.__name__,
            "generation",
            "materials",
            "zinc250k_v0",
            TorchDrugGenerator.__name__,
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            TorchDrugGraphAF.__name__,
            "generation",
            "materials",
            "zinc250k_v0",
            TorchDrugGenerator.__name__,
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            TorchDrugGCPN.__name__,
            "generation",
            "materials",
            "qed_v0",
            TorchDrugGenerator.__name__,
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            TorchDrugGraphAF.__name__,
            "generation",
            "materials",
            "qed_v0",
            TorchDrugGenerator.__name__,
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            TorchDrugGCPN.__name__,
            "generation",
            "materials",
            "plogp_v0",
            TorchDrugGenerator.__name__,
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            TorchDrugGraphAF.__name__,
            "generation",
            "materials",
            "plogp_v0",
            TorchDrugGenerator.__name__,
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
    ],
)
def test_generation_via_registry(
    algorithm_type: str,
    domain: str,
    algorithm_name: str,
    algorithm_version: str,
    algorithm_application,
):
    algorithm = ApplicationsRegistry.get_application_instance(
        target=None,
        algorithm_type=algorithm_type,
        domain=domain,
        algorithm_name=algorithm_name,
        algorithm_application=algorithm_application,
    )
    items = list(algorithm.sample(1))
    assert len(items) == 1



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Model builder module."""

import logging
from collections import OrderedDict
from typing import Any, Dict, List
from typing import OrderedDict as OrderedDictType

import torch

from ....torch import device_claim
from . import ARCHITECTURE_FACTORY
from .base_model import GranularBaseModel

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def build_model(architecture: Dict[str, Any]) -> GranularBaseModel:
    """Build model from architecture configuration.

    Args:
        architecture: architecture configuration.

    Returns:
        built model.
    """
    model_name = architecture["name"]
    model_type = architecture["type"].lower()
    if model_type not in ARCHITECTURE_FACTORY:
        raise ValueError(
            f"model_type={model_type} not supported. Pick a valid one: {sorted(ARCHITECTURE_FACTORY.keys())}"
        )
    model = ARCHITECTURE_FACTORY[model_type](
        data=architecture["data"], **architecture["hparams"]
    )

    if architecture["start_from_checkpoint"]:
        loaded_params = torch.load(
            architecture["hparams"]["checkpoint_path"], map_location=device_claim(None)
        )
        loaded_architecture_latent = loaded_params["hyper_parameters"][
            "architecture_latent_models"
        ]
        loaded_architecture_autoencoder = loaded_params["hyper_parameters"][
            "architecture_autoencoders"
        ]
        for arcihtecture_autoencoder in loaded_architecture_autoencoder:
            if model_name == arcihtecture_autoencoder["name"]:
                architecture = arcihtecture_autoencoder
        for architecture_latent in loaded_architecture_latent:
            if model_name == architecture_latent["name"]:
                architecture = architecture_latent
        loaded_state_dict: OrderedDictType[str, torch.Tensor] = OrderedDict()
        for layer_name in loaded_params["state_dict"]:
            state_model_name, *layer_name_elements = layer_name.split(".")
            state_name = ".".join(layer_name_elements)
            try:
                checkpoint_model_name = architecture["hparams"]["checkpoint_model_name"]
            except Exception:
                checkpoint_model_name = None
            if (
                state_model_name == model_name
                or state_model_name == checkpoint_model_name
            ):
                loaded_state_dict[state_name] = loaded_params["state_dict"][layer_name]
        model.load_state_dict(loaded_state_dict)
        model.name = model_name
        model.data = architecture["data"]
        model.target_key = model_name + "_" + architecture["data"]["target"]
        try:
            freeze_weights = architecture["freeze_weights"]
        except KeyError:
            freeze_weights = None

        if freeze_weights:
            for param in model.parameters():
                param.requires_grad = False
            model.eval()
        if model_type == "mlp_predictor":
            model.from_position = architecture["from_position"]
        else:
            model.position = architecture["position"]
            model.input_key = model_name + "_" + architecture["data"]["input"]
    return model


def building_models(architectures: List[Dict[str, Any]]) -> List[GranularBaseModel]:
    """Building models given architecture configurations.

    Args:
        architectures: list of architecture configurations.

    Returns:
        a list of models.
    """
    return [build_model(architecture) for architecture in architectures]


def define_latent_models_input_size(
    architecture_autoencoders: List[Dict[str, Any]],
    architecture_latent_models: List[Dict[str, Any]],
) -> List[Dict[str, Any]]:
    """Define latent models input size and return the updated configurations.

    Args:
        architecture_autoencoders: list of autoencoder architecture configurations.
        architecture_latent_models: list of latent model architecture configurations.

    Returns:
        list of update latent model architecture configurations.
    """
    size_autoencoder: Dict[str, int] = dict()
    for architecture in architecture_autoencoders:
        if architecture["position"] not in size_autoencoder.keys():
            size_autoencoder[architecture["position"]] = architecture["hparams"][
                "latent_size"
            ]
        else:
            logger.warning(f"position for architecture={architecture} is not unique!")

    updated_architecture_latent_models = []
    for _, architecture in enumerate(architecture_latent_models):
        architecture["hparams"]["input_size"] = sum(
            [size_autoencoder[pos] for pos in architecture["from_position"]]
        )
        updated_architecture_latent_models.append(architecture)

    return updated_architecture_latent_models



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import copy
from itertools import count
from typing import Any, Dict, List, Tuple, Union

import numpy as np
import torch
import torch.nn as nn
import torch_geometric.data as gd
from torch import Tensor
from torch_scatter import scatter

from ..envs.graph_building_env import (
    Graph,
    GraphAction,
    GraphActionCategorical,
    GraphActionType,
    GraphBuildingEnv,
    GraphBuildingEnvContext,
    generate_forward_trajectory,
)


class TrajectoryBalanceModel(nn.Module):
    """Generic model compatible with trajectory balance."""

    def forward(self, batch: gd.Batch) -> Tuple[GraphActionCategorical, Tensor]:
        """Run forward pass.

        Args:
            batch: batch of data

        Returns:
            action: action to take
        """
        raise NotImplementedError()

    def log_z(self, cond_info: Tensor) -> Tensor:
        """Compute log_z.

        Args:
            cond_info: conditional information

        Returns:
            log partition function.
        """
        raise NotImplementedError()


class TrajectoryBalance:
    """A trajectory balance algorithm for gflownet."""

    def __init__(
        self,
        configuration: Dict[str, Any],
        environment: GraphBuildingEnv,
        context: GraphBuildingEnvContext,
        max_len: int = None,
    ):
        """Initialize trajectory balance algorithm.

        Trajectory balance implementation, see
        "Trajectory Balance: Improved Credit Assignment in GFlowNets
            Nikolay Malkin, Moksh Jain, Emmanuel Bengio, Chen Sun, Yoshua Bengio"
            https://arxiv.org/abs/2201.13259.

        Code adapted from: https://github.com/recursionpharma/gflownet/blob/trunk/src/gflownet/algo/trajectory_balance.py.

        Args
            configuration: hyperparameters.
            environment: a graph environment.
            context: a context.
            rng: rng used to take random actions.
            hps: hyperparameter dictionary, see above for used keys.
                - random_action_prob: float, probability of taking a uniform random action when sampling.
                - illegal_action_logreward: float, log(R) given to the model for non-sane end states or illegal actions.
                - bootstrap_own_reward: bool, if True, uses the .reward batch data to predict rewards for sampled data.
                - tb_epsilon: float, if not None, adds this epsilon in the numerator and denominator of the log-ratio.
                - reward_loss_multiplier: float, multiplying constant for the bootstrap loss.
            max_len: if not None, ends trajectories of more than max_len steps.
            max_nodes: if not None, ends trajectories of graphs with more than max_nodes steps (illegal action).
        """

        self.ctx = context
        self.env = environment
        self.hps = configuration
        self.rng = self.hps["rng"]
        self.max_len = max_len
        self.max_nodes = self.hps["max_nodes"]
        self.random_action_prob = self.hps["random_action_prob"]
        self.illegal_action_logreward = self.hps["illegal_action_logreward"]
        self.bootstrap_own_reward = self.hps["bootstrap_own_reward"]
        self.sanitize_samples = True
        self.epsilon = self.hps["tb_epsilon"]
        self.reward_loss_multiplier = self.hps["reward_loss_multiplier"]
        # Experimental flags
        self.reward_loss_is_mae = True
        self.tb_loss_is_mae = False
        self.tb_loss_is_huber = False
        self.mask_invalid_rewards = False
        self.length_normalize_losses = False
        self.sample_temp = 1

    def _corrupt_actions(
        self, actions: List[Tuple[int, int, int]], cat: GraphActionCategorical
    ):
        """Sample from the uniform policy with probability random_action_prob.

        Args:
            actions: list of actions.
            cat: action categorical.
        """

        if self.random_action_prob <= 0:
            return
        (corrupted,) = (
            self.rng.uniform(size=len(actions)) < self.random_action_prob
        ).nonzero()
        for i in corrupted:
            n_in_batch = [int((b == i).sum()) for b in cat.batch]
            n_each = np.array(
                [
                    float(logit.shape[1]) * nb
                    for logit, nb in zip(cat.logits, n_in_batch)
                ]
            )
            which = self.rng.choice(len(n_each), p=n_each / n_each.sum())
            row = self.rng.choice(n_in_batch[which])
            col = self.rng.choice(cat.logits[which].shape[1])
            actions[i] = (which, row, col)

    def create_training_data_from_own_samples(
        self, model: Union[nn.Module, TrajectoryBalanceModel], n: int, cond_info: Tensor
    ) -> List[Dict]:
        """Generate trajectories by sampling a model.

        Args:
            model: model used with a certain algorithm (i.e. trajectory balance). The model being sampled.
            graphs: list of N Graph endpoints.
            cond_info: conditional information, shape (N, n_info).

        Returns:
            data: a list of trajectories. Each trajectory is a dict with keys.
                - trajs: List[Tuple[Graph, GraphAction]].
                - reward_pred: float, -100 if an illegal action is taken, predicted R(x) if bootstrapping, None otherwise.
                - fwd_logprob: log_z + sum logprobs P_F.
                - bck_logprob: sum logprobs P_B.
                - log_z: predicted log_z.
                - loss: predicted loss (if bootstrapping).
                - is_valid: is the generated graph valid according to the environment and context.
        """

        ctx = self.ctx
        env = self.env
        dev = self.ctx.device
        cond_info = cond_info.to(dev)
        # how do we compute log_z pred?
        log_z_pred = model.log_z(cond_info)  # type: ignore

        # This will be returned as training data
        data: List[Dict] = []
        for i in range(n):
            data.append({"traj": [], "reward_pred": None, "is_valid": True})

        # Let's also keep track of trajectory statistics according to the model
        zero = torch.tensor([0], device=dev).float()
        fwd_logprob: List[List[Tensor]] = [[] for i in range(n)]
        bck_logprob: List[List[Tensor]] = [
            [zero] for i in range(n)
        ]  # zero in case there is a single invalid action

        graphs = [env.new() for i in range(n)]
        done = [False] * n

        def not_done(lst):
            return [e for i, e in enumerate(lst) if not done[i]]

        # TODO report these stats:
        mol_too_big = 0
        mol_not_sane = 0
        invalid_act = 0
        logprob_of_illegal = []

        illegal_action_logreward = torch.tensor(
            [self.illegal_action_logreward], device=dev
        )
        if self.epsilon is not None:
            epsilon = torch.tensor([self.epsilon], device=dev).float()

        for t in range(self.max_len) if self.max_len is not None else count(0):

            # Construct graphs for the trajectories that aren't yet done
            torch_graphs = [ctx.graph_to_data(i) for i in not_done(graphs)]
            not_done_mask = torch.tensor(done, device=dev).logical_not()
            # Forward pass to get GraphActionCategorical
            fwd_cat, log_reward_preds = model(
                ctx.collate(torch_graphs).to(dev), cond_info[not_done_mask]
            )
            if self.sample_temp != 1:
                sample_cat = copy.copy(fwd_cat)
                sample_cat.logits = [i / self.sample_temp for i in fwd_cat.logits]
                actions = sample_cat.sample()
            else:
                actions = fwd_cat.sample()
            self._corrupt_actions(actions, fwd_cat)
            graph_actions = [
                ctx.aidx_to_graph_action(g, a) for g, a in zip(torch_graphs, actions)
            ]
            log_probs = fwd_cat.log_prob(actions)

            for i, j in zip(not_done(range(n)), range(n)):
                # Step each trajectory, and accumulate statistics
                fwd_logprob[i].append(log_probs[j].unsqueeze(0))
                data[i]["traj"].append((graphs[i], graph_actions[j]))
                # Check if we're done
                if graph_actions[j].action is GraphActionType.Stop:
                    done[i] = True
                    if self.sanitize_samples and not ctx.is_sane(graphs[i]):
                        # check if the graph is sane (e.g. RDKit can
                        # construct a molecule from it) otherwise
                        # treat the done action as illegal
                        mol_not_sane += 1
                        data[i]["reward_pred"] = illegal_action_logreward.exp()
                        data[i]["is_valid"] = False
                    elif self.bootstrap_own_reward:
                        # if we're bootstrapping, extract reward prediction
                        data[i]["reward_pred"] = log_reward_preds[j].detach().exp()
                else:  # If not done, try to step the environment
                    gp = graphs[i]
                    try:
                        # env.step can raise AssertionError if the action is illegal
                        gp = env.step(graphs[i], graph_actions[j])
                        if self.max_nodes is not None:
                            assert len(gp.nodes) <= self.max_nodes
                    except AssertionError:
                        if len(gp.nodes) > self.max_nodes:
                            mol_too_big += 1
                        else:
                            invalid_act += 1
                        done[i] = True
                        data[i]["reward_pred"] = illegal_action_logreward.exp()
                        data[i]["is_valid"] = False
                        continue
                    # Add to the trajectory
                    # P_B = uniform backward
                    n_back = env.count_backward_transitions(gp)
                    bck_logprob[i].append(torch.tensor([1 / n_back], device=dev).log())
                    graphs[i] = gp
            if all(done):
                break

        for i in range(n):
            # If we're not bootstrapping, we could query the reward
            # model here, but this is expensive/impractical.  Instead
            # just report forward and backward flows
            data[i]["log_z"] = log_z_pred[i].item()
            data[i]["fwd_logprob"] = sum(fwd_logprob[i])
            data[i]["bck_logprob"] = sum(bck_logprob[i])
            if self.bootstrap_own_reward:
                if not data[i]["is_valid"]:
                    logprob_of_illegal.append(data[i]["fwd_logprob"].item())
                # If we are bootstrapping, we can report the theoretical loss as well
                numerator = data[i]["fwd_logprob"] + log_z_pred[i]
                denominator = data[i]["bck_logprob"] + data[i]["reward_pred"].log()
                if self.epsilon is not None:
                    numerator = torch.logaddexp(numerator, epsilon)
                    denominator = torch.logaddexp(denominator, epsilon)
                data[i]["loss"] = (numerator - denominator).pow(2)
        return data

    def create_training_data_from_graphs(
        self, graphs: List[Graph]
    ) -> List[Dict[str, List[Tuple[Graph, GraphAction]]]]:
        """Generate trajectories from known endpoints.

        Args:
            graphs: list of Graph endpoints.

        Returns:
            a list of trajectories.
        """
        return [{"traj": generate_forward_trajectory(i)} for i in graphs]

    def construct_batch(
        self,
        trajs: List[Dict[str, List[Tuple[Graph, GraphAction]]]],
        cond_info: float,
        rewards: float,
    ) -> gd.Batch:
        """Construct a batch from a list of trajectories and their information.

        Args:
            trajs: a list of N trajectories.
            cond_info: the conditional info that is considered for each trajectory. Shape (N, n_info).
            rewards: the transformed reward (e.g. R(x) ** beta) for each trajectory. Shape (N,).

        Returns:
            a (CPU) Batch object with relevant attributes added.
        """

        torch_graphs = [
            self.ctx.graph_to_data(i[0]) for tj in trajs for i in tj["traj"]
        ]
        actions = [
            self.ctx.graph_action_to_aidx(g, a)
            for g, a in zip(torch_graphs, [i[1] for tj in trajs for i in tj["traj"]])
        ]
        num_backward = torch.tensor(
            [
                # Count the number of backward transitions from s_{t+1},
                # unless t+1 = T is the last time step
                self.env.count_backward_transitions(tj["traj"][i + 1][0])
                if i + 1 < len(tj["traj"])
                else 1
                for tj in trajs
                for i in range(len(tj["traj"]))
            ]
        )

        batch = self.ctx.collate(torch_graphs)
        batch.traj_lens = torch.tensor([len(i["traj"]) for i in trajs])
        batch.num_backward = num_backward
        batch.actions = torch.tensor(actions)
        batch.rewards = rewards
        batch.cond_info = cond_info
        batch.is_valid = torch.tensor([i.get("is_valid", True) for i in trajs]).float()
        return batch

    def compute_batch_losses(
        self, model: TrajectoryBalanceModel, batch: gd.Batch, num_bootstrap: int = 0
    ) -> Tuple[torch.Tensor, Dict]:
        """Compute the losses over trajectories contained in the batch.

        Args:
            model: a GNN taking in a batch of graphs as input as per constructed by self.construct_batch.
                    Must have a `log_z` attribute, itself a model, which predicts log of z(cond_info)
            batch: batch of graphs inputs as per constructed by self.construct_batch.
            num_bootstrap: the number of trajectories for which the reward loss is computed. Ignored if 0.

        Returns:
            a tuple containing the loss for each trajectory and relevant info.
        """

        dev = batch.x.device
        # A single trajectory is comprised of many graphs
        num_trajs = int(batch.traj_lens.shape[0])
        rewards = batch.rewards
        cond_info = batch.cond_info

        # This index says which trajectory each graph belongs to, so
        # it will look like [0,0,0,0,1,1,1,2,...] if trajectory 0 is
        # of length 4, trajectory 1 of length 3, and so on.
        batch_idx = torch.arange(num_trajs, device=dev).repeat_interleave(
            batch.traj_lens
        )
        # The position of the last graph of each trajectory
        final_graph_idx = torch.cumsum(batch.traj_lens, 0) - 1

        # Forward pass of the model, returns a GraphActionCategorical and the optional bootstrap predictions
        fwd_cat, log_reward_preds = model(batch, cond_info[batch_idx])

        # Retreive the reward predictions for the full graphs,
        # i.e. the final graph of each trajectory
        log_reward_preds = log_reward_preds[final_graph_idx, 0]
        # Compute trajectory balance objective
        logz = model.log_z(cond_info)[:, 0]
        # This is the log prob of each action in the trajectory
        log_prob = fwd_cat.log_prob(batch.actions)
        # The log prob of each backward action
        log_p_B = (1 / batch.num_backward).log()
        # Take log rewards, and clip
        Rp = torch.maximum(rewards.log(), torch.tensor(-100.0, device=dev))
        # This is the log probability of each trajectory
        traj_log_prob = scatter(
            log_prob, batch_idx, dim=0, dim_size=num_trajs, reduce="sum"
        )
        # Compute log numerator and denominator of the TB objective
        numerator = logz + traj_log_prob
        denominator = Rp + scatter(
            log_p_B, batch_idx, dim=0, dim_size=num_trajs, reduce="sum"
        )

        if self.epsilon is not None:
            # Numerical stability epsilon
            epsilon = torch.tensor([self.epsilon], device=dev).float()
            numerator = torch.logaddexp(numerator, epsilon)
            denominator = torch.logaddexp(denominator, epsilon)

        invalid_mask = 1 - batch.is_valid
        if self.mask_invalid_rewards:
            # Instead of being rude to the model and giving a
            # logreward of -100 what if we say, whatever you think the
            # logprobablity of this trajetcory is it should be smaller
            # (thus the `numerator - 1`). Why 1? Intuition?
            denominator = denominator * (1 - invalid_mask) + invalid_mask * (
                numerator.detach() - 1
            )

        if self.tb_loss_is_mae:
            traj_losses = abs(numerator - denominator)
        elif self.tb_loss_is_huber:
            pass  # TODO
        else:
            traj_losses = (numerator - denominator).pow(2)

        # Normalize losses by trajectory length
        if self.length_normalize_losses:
            traj_losses = traj_losses / batch.traj_lens

        if self.bootstrap_own_reward:
            num_bootstrap = num_bootstrap or len(rewards)
            if self.reward_loss_is_mae:
                reward_losses = abs(
                    rewards[:num_bootstrap] - log_reward_preds[:num_bootstrap].exp()
                )
            else:
                reward_losses = (
                    rewards[:num_bootstrap] - log_reward_preds[:num_bootstrap].exp()
                ).pow(2)
            reward_loss = reward_losses.mean()
        else:
            reward_loss = 0

        loss = traj_losses.mean() + reward_loss * self.reward_loss_multiplier
        info = {
            "offline_loss": traj_losses[: batch.num_offline].mean(),
            "online_loss": traj_losses[batch.num_offline :].mean(),
            "reward_loss": reward_loss,
            "invalid_trajectories": invalid_mask.mean() * 2,
            "invalid_logprob": (invalid_mask * traj_log_prob).sum()
            / (invalid_mask.sum() + 1e-4),
            "invalid_losses": (invalid_mask * traj_losses).sum()
            / (invalid_mask.sum() + 1e-4),
            "log_z": logz.mean(),
        }

        if not torch.isfinite(traj_losses).all():
            raise ValueError("loss is not finite")
        return loss, info



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import ast
import os
import pickle
from typing import Any, Callable, Dict, List, NewType, Tuple, Union

import numpy as np
import pandas as pd
import rdkit.Chem as Chem
import torch
import torch.nn as nn
import torch_geometric.data as gd
from torch import Tensor
from torch.utils.data import Dataset

from ..envs.graph_building_env import GraphActionCategorical
from ..util import MPModelPlaceholder, wrap_model_mp

# This type represents an unprocessed list of reward signals/conditioning information
FlatRewards = NewType("FlatRewards", torch.tensor)  # type: ignore

# This type represents the outcome for a multi-objective task of
# converting FlatRewards to a scalar, e.g. (sum R_i omega_i) ** beta
RewardScalar = NewType("RewardScalar", torch.tensor)  # type: ignore


class GFlowNetDataset(Dataset):
    """A dataset for gflownet."""

    def __init__(
        self, h5_file: str = None, target: str = "gap", properties: List[str] = []
    ) -> None:

        """Initialize a gflownet dataset.
        If the dataset is in a format compatible with h5 file, we can directly load it.
        If the dataset is in a format compatible with xyz file, we have to convert it to h5 file.

        Code adapted from: https://github.com/recursionpharma/gflownet/tree/trunk/src/gflownet/data.

        Args:
            h5_file: data file in h5 format.
            target: reward target.
            properties: relevant properties for the task.

        Raises:
            ValueError: if the dataset is not in a format compatible with h5 file or not present.
        """
        if h5_file:
            try:
                # pickle on py3.8 uses prot 5
                pickle.HIGHEST_PROTOCOL = 5
                self.df = pd.HDFStore(h5_file, "r")["df"]
            except FileNotFoundError:
                ValueError("No h5 compatible database found in the path.")
        else:
            raise ValueError("The h5_file path is None. Please specify a h5 file path.")

        self.target = target
        self.properties = properties

    def set_indexes(self, ixs: torch.Tensor):
        """Set the indexes of the dataset split (train/val/test).

        Args:
            ixs: indexes of the dataset split.
        """
        self.idcs = ixs

    def get_len(self):
        """Get the length of the full dataset (before splitting).

        Returns:
            len: the length of the full dataset.
        """
        return len(self.df)

    def get_stats(self, percentile: float = 0.95) -> Tuple[float, float, Any]:
        """Get the stats of the dataset.

        Args:
            percentile: percentile.

        Returns:
            min, max, percentile.
        """
        y = self.df[self.target].astype(float)
        return y.min(), y.max(), np.sort(y)[int(y.shape[0] * percentile)]

    @staticmethod
    def convert_xyz_to_h5(
        xyz_path: str = "data/xyz",
        h5_path: str = "data/qm9.h5",
        property_names: List[str] = [
            "rA",
            "rB",
            "rC",
            "mu",
            "alpha",
            "homo",
            "lumo",
            "gap",
            "r2",
            "zpve",
            "U0",
            "U",
            "H",
            "G",
            "Cv",
        ],
    ) -> None:
        """Convert the data from xyz to h5. Assumes that the xyz files are extracted in the xyz_path folder.

        Args:
            xyz_path: path to the xyz file in input.
            h5_path: path to the h5 file in output.
            property_names: names of the properties we want to use/overwrite.
        """

        # Reads the xyz files and return the properties, smiles and coordinates
        data = []
        smiles = []
        properties = []
        i = 0

        for file in os.listdir(xyz_path):
            try:
                path = os.path.join(xyz_path, file)
                atoms, coordinates, smile, prop = GFlowNetDataset._read_xyz(path)
                data.append(
                    (atoms, coordinates)
                )  # A tuple with the atoms and its coordinates
                smiles.append(smile)  # The SMILES representation
                properties.append(prop)  # The molecules properties
            except ValueError:
                print(path)
            i += 1

        # rename relevant properties to match GFN schema
        labels = property_names
        df = pd.DataFrame(properties, columns=labels)
        df["SMILES"] = smiles

        df.to_hdf(h5_path, key="df", mode="w")

    @staticmethod
    def _read_xyz(path: str):
        """Reads the xyz files in the directory on 'path'.

        Code adapted from: https://www.kaggle.com/code/rmonge/predicting-molecule-properties-based-on-its-smiles/notebook

        Args:
            path: the path to the folder.

        Returns:
            atoms: list with the characters representing the atoms of a molecule.
            coordinates: list with the cartesian coordinates of each atom.
            smile: list with the SMILE representation of a molecule.
            prop: list with the scalar properties.
        """
        atoms = []
        coordinates = []

        with open(path, "r") as file:
            lines = file.readlines()
            n_atoms = int(lines[0])  # the number of atoms
            smile = lines[n_atoms + 3].split()[0]  # smiles string
            prop = lines[1].split()[2:]  # scalar properties

            # to retrieve each atmos and its cartesian coordenates
            for atom in lines[2 : n_atoms + 2]:
                line = atom.split()
                # which atom
                atoms.append(line[0])

                # its coordinate
                # Some properties have '*^' indicading exponentiation
                try:
                    coordinates.append((float(line[1]), float(line[2]), float(line[3])))
                except ValueError:
                    coordinates.append(
                        (
                            float(line[1].replace("*^", "e")),
                            float(line[2].replace("*^", "e")),
                            float(line[3].replace("*^", "e")),
                        )
                    )

        return atoms, coordinates, smile, prop

    def __len__(self):
        """Dataset split (train/val/test) length.

        Returns:
            length of the dataset.
        """
        return len(self.idcs)

    def __getitem__(self, idx: torch.Tensor) -> Tuple[Any, float]:
        """Retrieve an item from the dataset by index.

        Args:
            index: index for the item.

        Returns:
            an tuple (item, reward).
        """
        return (
            Chem.MolFromSmiles(self.df["SMILES"][self.idcs[idx]]),
            self.df[self.target][self.idcs[idx]],
        )


class GFlowNetTask:
    """A task for gflownet."""

    def __init__(
        self,
        configuration: Dict[str, Any],
        dataset: GFlowNetDataset,
        reward_model: nn.Module = None,
        wrap_model: Callable[[nn.Module], nn.Module] = None,
    ) -> None:

        """Initialize a generic gflownet task.
        The task specifies the reward model for the trajectory.
        We consider the task as part of the dataset.

        Code adapted from: https://github.com/recursionpharma/gflownet/tree/trunk/src/gflownet/tasks.

        Args:
            configuration: a dictionary with the task configuration.
            dataset: a dataset instance.
            reward_model: The model that is used to generate the conditional reward.
            wrap_model: a wrapper function that is applied to the model.
        """
        hps = configuration

        self._wrap_model = wrap_model
        self.num_workers = hps["num_workers"]
        self.device = hps["device"]
        self.rng = hps["rng"]

        # if reward model provided, use it
        if reward_model:
            self.model = {"model_task": reward_model}
        else:
            self.model = self.load_task_models()

        self.dataset = dataset
        self.temperature_sample_dist = hps["temperature_sample_dist"]
        self.temperature_dist_params = ast.literal_eval(hps["temperature_dist_params"])

        self._min, self._max, self._percentile_95 = self.dataset.get_stats(percentile=0.05)  # type: ignore
        self._width = self._max - self._min
        self._rtrans = "unit+95p"

    def load_task_models(self) -> Dict[str, nn.Module]:
        """Loads the task models.

        Returns:
            model: a dictionary with the task models.
        """
        pass

    def sample_conditional_information(self, n: int) -> Dict[str, Any]:
        """Samples conditional information for a minibatch.

        Args:
            n: number of samples.

        Returns:
            cond_info: a dictionary with the sampled conditional information.
        """
        pass

    def cond_info_to_reward(
        self, cond_info: Dict[str, Any], flat_reward: FlatRewards
    ) -> RewardScalar:
        """Combines a minibatch of reward signal vectors and conditional information into a scalar reward.

        Args:
            cond_info: a dictionary with various conditional informations (e.g. temperature).
            flat_reward: a 2d tensor where each row represents a series of flat rewards.

        Returns:
            reward: a 1d tensor, a scalar reward for each minibatch entry.
        """
        raise NotImplementedError()

    def compute_flat_rewards(self, x: List[Any]) -> Tuple[RewardScalar, Any]:
        """Compute the flat rewards of mols according the the tasks' proxies.

        Args:
            mols: a list of RDKit molecules.
        Returns:
            reward: a 1d tensor, a scalar reward for each molecule.
            is_valid: a 1d tensor, a boolean indicating whether the molecule is valid.
        """
        raise NotImplementedError()

    def flat_reward_transform(self, y: Union[float, Tensor]) -> FlatRewards:
        """Transforms a reward with a generic structure to a flat vector.

        Args:
            y: scalar reward for a trajectory.
        """
        raise NotImplementedError()

    def _wrap_model_mp(self, model: nn.Module) -> Union[nn.Module, MPModelPlaceholder]:
        """Wraps a nn.Module instance so that it can be shared to `DataLoader` workers.

        Args:
            model: a nn.Module instance.
        """
        if self.num_workers > 0:
            placeholder = wrap_model_mp(
                model, self.num_workers, cast_types=(gd.Batch, GraphActionCategorical)
            )
            return placeholder
        return model



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Moses VAE trainer unit tests."""
import os
import shutil
import tempfile
from typing import Any, Dict, cast

from gt4sd.training_pipelines import TRAINING_PIPELINE_MAPPING, MosesVAETrainingPipeline

from ...tests.utils import exitclose_file_creator

DATA_PATH = exitclose_file_creator("training_pipelines/tests/molecules.smiles")


def _create_training_output_filepaths(directory: str) -> Dict[str, str]:
    """Create output filepath from directory.

    Args:
        directory: output directory.

    Returns:
        a dictionary containing the output files.
    """
    return {
        "log_file": os.path.join(directory, "log.txt"),
        "model_save": os.path.join(directory, "model.pt"),
        "config_save": os.path.join(directory, "config.pt"),
        "vocab_save": os.path.join(directory, "vocab.pt"),
    }


template_config = {
    "model_args": {
        "embedding_size": 32,
        "q_cell": "gru",
        "q_bidir": False,
        "q_d_h": 256,
        "q_n_layers": 1,
        "q_dropout": 0.5,
        "d_cell": "gru",
        "d_n_layers": 3,
        "d_dropout": 0,
        "d_d_h": 512,
        "d_z": 128,
        "freeze_embeddings": False,
    },
    "training_args": {
        "n_batch": 512,
        "grad_clipping": 50,
        "kl_start": 0,
        "kl_w_start": 0,
        "kl_w_end": 0.05,
        "lr_start": 3 * 1e-4,
        "lr_n_period": 1,
        "lr_n_restarts": 1,
        "lr_n_mult": 1,
        "lr_end": 3 * 1e-4,
        "n_last": 1000,
        "n_jobs": 1,
        "n_workers": 1,
        "save_frequency": 1,
        "seed": 0,
        "device": "cpu",
        "save_frequency": 1,
        "warm_start": "",
    },
    "dataset_args": {"train_load": DATA_PATH, "val_load": DATA_PATH},
}


def test_train():

    pipeline = TRAINING_PIPELINE_MAPPING.get("moses-vae-trainer")

    assert pipeline is not None

    TEMPORARY_DIRECTORY = tempfile.mkdtemp()

    test_pipeline = cast(MosesVAETrainingPipeline, pipeline())

    config: Dict[str, Any] = template_config.copy()
    for key, value in _create_training_output_filepaths(TEMPORARY_DIRECTORY).items():
        config["training_args"][key] = value

    test_pipeline.train(**config)

    shutil.rmtree(TEMPORARY_DIRECTORY)



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Model module."""

from __future__ import division, print_function

from typing import Any, Dict

import torch
import torch.nn as nn


class ConvLayer(nn.Module):
    """Convolutional operation on graphs."""

    def __init__(self, atom_fea_len: int, nbr_fea_len: int):
        """Initialize ConvLayer.

        Args:
            atom_fea_len: int
              Number of atom hidden features.
            nbr_fea_len: int
              Number of bond features.
        """
        super(ConvLayer, self).__init__()
        self.atom_fea_len = atom_fea_len
        self.nbr_fea_len = nbr_fea_len
        self.fc_full = nn.Linear(
            2 * self.atom_fea_len + self.nbr_fea_len, 2 * self.atom_fea_len
        )
        self.sigmoid = nn.Sigmoid()
        self.softplus1 = nn.Softplus()
        self.bn1 = nn.BatchNorm1d(2 * self.atom_fea_len)
        self.bn2 = nn.BatchNorm1d(self.atom_fea_len)
        self.softplus2 = nn.Softplus()

    def forward(
        self,
        atom_in_fea: torch.Tensor,
        nbr_fea: torch.Tensor,
        nbr_fea_idx: torch.LongTensor,
    ) -> torch.Tensor:
        """Forward pass.

        N: Total number of atoms in the batch.
        M: Max number of neighbors.

        Args:
            atom_in_fea: Variable(torch.Tensor) shape (N, atom_fea_len)
              Atom hidden features before convolution.
            nbr_fea: Variable(torch.Tensor) shape (N, M, nbr_fea_len)
              Bond features of each atom's M neighbors.
            nbr_fea_idx: torch.LongTensor shape (N, M)
              Indices of M neighbors of each atom.

        Returns:
            atom_out_fea: nn.Variable shape (N, atom_fea_len)
              Atom hidden features after convolution.

        """
        # TODO will there be problems with the index zero padding?
        N, M = nbr_fea_idx.shape
        # convolution
        atom_nbr_fea = atom_in_fea[nbr_fea_idx, :]
        total_nbr_fea = torch.cat(
            [
                atom_in_fea.unsqueeze(1).expand(N, M, self.atom_fea_len),
                atom_nbr_fea,
                nbr_fea,
            ],
            dim=2,
        )
        total_gated_fea = self.fc_full(total_nbr_fea)
        total_gated_fea = self.bn1(
            total_gated_fea.view(-1, self.atom_fea_len * 2)
        ).view(N, M, self.atom_fea_len * 2)
        nbr_filter, nbr_core = total_gated_fea.chunk(2, dim=2)
        nbr_filter = self.sigmoid(nbr_filter)
        nbr_core = self.softplus1(nbr_core)
        nbr_sumed = torch.sum(nbr_filter * nbr_core, dim=1)
        nbr_sumed = self.bn2(nbr_sumed)
        out = self.softplus2(atom_in_fea + nbr_sumed)
        return out


class CrystalGraphConvNet(nn.Module):
    """Create a crystal graph convolutional neural network for predicting total material properties."""

    def __init__(
        self,
        orig_atom_fea_len: int,
        nbr_fea_len: int,
        atom_fea_len: int = 64,
        n_conv: int = 3,
        h_fea_len: int = 128,
        n_h: int = 1,
        classification: bool = False,
    ):
        """Initialize CrystalGraphConvNet.

        Args:
            orig_atom_fea_len: int
              Number of atom features in the input.
            nbr_fea_len: int
              Number of bond features.
            atom_fea_len: int
              Number of hidden atom features in the convolutional layers.
            n_conv: int
              Number of convolutional layers.
            h_fea_len: int
              Number of hidden features after pooling.
            n_h: int
              Number of hidden layers after pooling.
        """
        super(CrystalGraphConvNet, self).__init__()
        self.classification = classification
        self.embedding = nn.Linear(orig_atom_fea_len, atom_fea_len)
        self.convs = nn.ModuleList(
            [
                ConvLayer(atom_fea_len=atom_fea_len, nbr_fea_len=nbr_fea_len)
                for _ in range(n_conv)
            ]
        )
        self.conv_to_fc = nn.Linear(atom_fea_len, h_fea_len)
        self.conv_to_fc_softplus = nn.Softplus()
        if n_h > 1:
            self.fcs = nn.ModuleList(
                [nn.Linear(h_fea_len, h_fea_len) for _ in range(n_h - 1)]
            )
            self.softpluses = nn.ModuleList([nn.Softplus() for _ in range(n_h - 1)])
        if self.classification:
            self.fc_out = nn.Linear(h_fea_len, 2)
        else:
            self.fc_out = nn.Linear(h_fea_len, 1)
        if self.classification:
            self.logsoftmax = nn.LogSoftmax(dim=1)
            self.dropout = nn.Dropout()

    def forward(
        self,
        atom_fea: torch.Tensor,
        nbr_fea: torch.Tensor,
        nbr_fea_idx: torch.LongTensor,
        crystal_atom_idx: torch.LongTensor,
    ) -> torch.Tensor:
        """Forward pass.

        N: Total number of atoms in the batch.
        M: Max number of neighbors.
        N0: Total number of crystals in the batch.

        Args:
            atom_fea: Variable(torch.Tensor) shape (N, orig_atom_fea_len)
              Atom features from atom type.
            nbr_fea: Variable(torch.Tensor) shape (N, M, nbr_fea_len)
              Bond features of each atom's M neighbors.
            nbr_fea_idx: torch.LongTensor shape (N, M)
              Indices of M neighbors of each atom.
            crystal_atom_idx: list of torch.LongTensor of length N0
              Mapping from the crystal idx to atom idx.

        Returns:
            prediction: nn.Variable shape (N, )
              Atom hidden features after convolution.
        """
        atom_fea = self.embedding(atom_fea)
        for conv_func in self.convs:
            atom_fea = conv_func(atom_fea, nbr_fea, nbr_fea_idx)
        crys_fea = self.pooling(atom_fea, crystal_atom_idx)
        crys_fea = self.conv_to_fc(self.conv_to_fc_softplus(crys_fea))
        crys_fea = self.conv_to_fc_softplus(crys_fea)
        if self.classification:
            crys_fea = self.dropout(crys_fea)
        if hasattr(self, "fcs") and hasattr(self, "softpluses"):
            for fc, softplus in zip(self.fcs, self.softpluses):
                crys_fea = softplus(fc(crys_fea))
        out = self.fc_out(crys_fea)
        if self.classification:
            out = self.logsoftmax(out)
        return out

    def pooling(
        self, atom_fea: torch.Tensor, crystal_atom_idx: torch.LongTensor
    ) -> torch.Tensor:
        """Pooling the atom features to crystal features.

        N: Total number of atoms in the batch.
        N0: Total number of crystals in the batch.

        Args:
            atom_fea: Variable(torch.Tensor) shape (N, atom_fea_len)
              Atom feature vectors of the batch.
            crystal_atom_idx: list of torch.LongTensor of length N0
              Mapping from the crystal idx to atom idx.
        """
        assert (
            sum([len(idx_map) for idx_map in crystal_atom_idx])
            == atom_fea.data.shape[0]
        )
        summed_fea = [
            torch.mean(atom_fea[idx_map], dim=0, keepdim=True)
            for idx_map in crystal_atom_idx
        ]
        return torch.cat(summed_fea, dim=0)


class Normalizer:
    """Normalize a Tensor and restore it later."""

    def __init__(self, tensor: torch.Tensor):
        """tensor is taken as a sample to calculate the mean and std."""
        self.mean = torch.mean(tensor)
        self.std = torch.std(tensor)

    def norm(self, tensor: torch.Tensor) -> torch.Tensor:
        """Noramlize a tensor.

        Args:
            tensor: tensor to be normalized.

        Returns:
            normalized tensor.
        """
        return (tensor - self.mean) / self.std

    def denorm(self, normed_tensor: torch.Tensor) -> torch.Tensor:
        """Denormalized tensor.

        Args:
            tensor: tensor to be denormalized:

        Returns:
            denormalized tensor.
        """
        return normed_tensor * self.std + self.mean

    def state_dict(self) -> Dict[str, torch.Tensor]:
        """Return the state dict of normalizer.

        Returns:
            dictionary including the used mean and std values.
        """
        return {"mean": self.mean, "std": self.std}

    def load_state_dict(self, state_dict: Dict[str, Any]) -> None:
        """Return the state dict of normalizer.

        Args:
            mean: mean value to be used for the normalization.
            std: std value to be used for the normalization.
        """
        self.mean = state_dict["mean"]
        self.std = state_dict["std"]



#!/usr/bin/env python
#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Run training pipelines for the GT4SD."""


import logging
import sys
from dataclasses import dataclass, field
from typing import IO, Iterable, Optional, Tuple, cast

from ..configuration import GT4SDConfiguration
from ..training_pipelines import (
    TRAINING_PIPELINE_ARGUMENTS_MAPPING,
    TRAINING_PIPELINE_MAPPING,
)
from ..training_pipelines.core import TrainingPipelineArguments
from .argument_parser import ArgumentParser, DataClass, DataClassType

logger = logging.getLogger(__name__)

SUPPORTED_TRAINING_PIPELINES = sorted(
    list(set(TRAINING_PIPELINE_ARGUMENTS_MAPPING) & set(TRAINING_PIPELINE_MAPPING))
)

# disable cudnn if issues with gpu training
if GT4SDConfiguration.get_instance().gt4sd_disable_cudnn:
    import torch

    torch.backends.cudnn.enabled = False


@dataclass
class TrainerArguments:
    """Trainer arguments."""

    __name__ = "trainer_base_args"

    training_pipeline_name: str = field(
        metadata={
            "help": f"Training pipeline name, supported pipelines: {', '.join(SUPPORTED_TRAINING_PIPELINES)}."
        },
    )
    configuration_file: Optional[str] = field(
        default=None,
        metadata={
            "help": "Configuration file for the training in JSON format. It can be used to completely by-pass pipeline specific arguments."
        },
    )


class TrainerArgumentParser(ArgumentParser):
    """Argument parser using a custom help logic."""

    def print_help(self, file: Optional[IO[str]] = None) -> None:
        """Print help checking dynamically whether a specific pipeline is passed.

        Args:
            file: an optional I/O stream. Defaults to None, a.k.a., stdout and stderr.
        """
        try:
            help_args_set = {"-h", "--help"}
            if (
                len(set(sys.argv).union(help_args_set)) < len(help_args_set) + 2
            ):  # considering filename
                super().print_help()
                return
            args = [arg for arg in sys.argv if arg not in help_args_set]
            parsed_arguments = super().parse_args_into_dataclasses(
                args=args, return_remaining_strings=True
            )
            trainer_arguments = None
            for arguments in parsed_arguments:
                if arguments.__name__ == "trainer_base_args":
                    trainer_arguments = arguments
                    break
            if trainer_arguments:
                trainer_arguments.training_pipeline_name
                training_pipeline_arguments = TRAINING_PIPELINE_ARGUMENTS_MAPPING.get(
                    trainer_arguments.training_pipeline_name, TrainingPipelineArguments
                )
                parser = ArgumentParser(
                    tuple(
                        [TrainerArguments, *training_pipeline_arguments]  # type:ignore
                    )
                )
                parser.print_help()
        except Exception:
            super().print_help()

    def parse_json_file(self, json_file: str) -> Tuple[DataClass, ...]:  # type: ignore
        """Overriding default .json parser.

        It by-passes all command line arguments and simply add the training pipeline.

        Args:
            json_file: JSON file containing pipeline configuration parameters.

        Returns:
            parsed arguments in a tuple of dataclasses.
        """
        number_of_dataclass_types = len(self.dataclass_types)  # type:ignore
        self.dataclass_types = [
            dataclass_type
            for dataclass_type in self.dataclass_types  # type:ignore
            if "gt4sd.cli.trainer.TrainerArguments" not in str(dataclass_type)
        ]
        try:
            parsed_arguments = super().parse_json_file(  # type:ignore
                json_file=json_file, allow_extra_keys=True
            )
        except Exception:
            logger.exception(
                f"error parsing configuration file: {json_file}, printing error and exiting"
            )
            sys.exit(1)
        if number_of_dataclass_types > len(self.dataclass_types):
            self.dataclass_types.insert(0, cast(DataClassType, TrainerArguments))
        return parsed_arguments


def main() -> None:
    """
    Run a training pipeline.

    Raises:
        ValueError: in case the provided training pipeline provided is not supported.
    """
    logging.basicConfig(stream=sys.stdout, level=logging.INFO)

    base_args = TrainerArgumentParser(
        cast(DataClassType, TrainerArguments)
    ).parse_args_into_dataclasses(return_remaining_strings=True)[0]
    training_pipeline_name = base_args.training_pipeline_name
    if training_pipeline_name not in set(SUPPORTED_TRAINING_PIPELINES):
        ValueError(
            f"Training pipeline {training_pipeline_name} is not supported. Supported types: {', '.join(SUPPORTED_TRAINING_PIPELINES)}."
        )
    arguments = TRAINING_PIPELINE_ARGUMENTS_MAPPING[training_pipeline_name]
    parser = TrainerArgumentParser(
        cast(Iterable[DataClassType], tuple([TrainerArguments, *arguments]))
    )

    configuration_filepath = base_args.configuration_file
    if configuration_filepath:
        args = parser.parse_json_file(json_file=configuration_filepath)

    else:
        args = parser.parse_args_into_dataclasses(return_remaining_strings=True)
    config = {
        arg.__name__: arg.__dict__
        for arg in args
        if isinstance(arg, TrainingPipelineArguments) and isinstance(arg.__name__, str)
    }

    if (
        base_args.training_pipeline_name == "granular-trainer"
        and config["model_args"]["model_list_path"] is None
    ):
        config["model_args"]["model_list_path"] = configuration_filepath

    pipeline = TRAINING_PIPELINE_MAPPING[training_pipeline_name]
    pipeline().train(**config)


if __name__ == "__main__":
    main()



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from typing import Callable, List, Optional, Union

from torchdrug import data

# isort: off
from torch import nn

"""
Necessary because torchdrug silently overwrites the default nn.Module. This is quite
invasive and causes significant side-effects in the rest of the code.
See: https://github.com/DeepGraphLearning/torchdrug/issues/77
"""
nn.Module = nn._Module  # type: ignore


class TorchDrugDataset(data.MoleculeDataset):
    """A generic TorchDrug dataset class that can be fed with custom data"""

    def __init__(
        self,
        file_path: str,
        target_fields: str,
        smiles_field: str = "smiles",
        verbose: int = 1,
        lazy: Optional[bool] = False,
        transform: Optional[Callable] = None,
        node_feature: Optional[Union[str, List[str]]] = "default",
        edge_feature: Optional[Union[str, List[str]]] = "default",
        graph_feature: Optional[Union[str, List[str]]] = None,
        with_hydrogen: Optional[bool] = False,
        kekulize: Optional[bool] = False,
    ):
        """
        Constructor of TorchDrugDataset.

        Args:
            file_path: The path to the .csv file containing the data.
            target_fields: The columns containing the property to be optimized.
            smiles_field: The column name containing the SMILES. Defaults to 'smiles'.
            verbose: output verbose level. Defaults to 1.
            lazy: If yes, molecules are processed in the dataloader. This is faster
                for setup, but slower at training time. Defaults to False.
            transform: Optional data transformation function. Defaults to None.
            node_feature: Node features to extract. Defaults to 'default'.
            edge_feature: Edge features to extract. Defaults to 'default'.
            graph_feature: Graph features to extract. Defaults to None.
            with_hydrogen: Whether hydrogens are stored in molecular graph.
                Defaults to False.
            kekulize: Whether aromatic bonds are converted to single/double bonds.
                Defaults to False.
        """

        self.load_csv(
            file_path,
            smiles_field=smiles_field,
            target_fields=target_fields,
            verbose=verbose,
            lazy=lazy,
            transform=transform,
            node_feature=node_feature,
            edge_feature=edge_feature,
            graph_feature=graph_feature,
            with_hydrogen=with_hydrogen,
            kekulize=kekulize,
        )



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Graph MCTS implementation."""

import os

from guacamol_baselines.graph_mcts import goal_directed_generation
from guacamol_baselines.graph_mcts.goal_directed_generation import GB_MCTS_Generator


class GraphMCTS:
    def __init__(
        self,
        init_smiles: str,
        population_size: int,
        n_jobs: int,
        generations: int,
        patience: int,
        num_sims: float,
        max_children: int,
        max_atoms: int,
        pickle_directory: str = os.path.dirname(
            os.path.realpath(goal_directed_generation.__file__)
        ),
    ):
        """Initialize SMILESGA.

        Args:
            init_smiles: path where to load hypothesis, candidate labels and, optionally, the smiles file.
            population_size: used with n_mutations for the initial generation of smiles within the population.
            n_jobs: number of concurrently running jobs.
            generations: number of evolutionary generations.
            patience: used for early stopping if population scores remains the same after generating molecules.
            num_sims: number of times to traverse the tree.
            max_children: maximum number of childerns a node could have.
            max_atoms: maximum number of atoms to explore to terminal the node state.
            pickle_directory: path from where to load pickle files.
        """
        self.init_smiles = init_smiles
        self.pickle_directory = pickle_directory
        self.population_size = population_size
        self.max_children = max_children
        self.n_jobs = n_jobs
        self.num_sims = num_sims
        self.generations = generations
        self.patience = patience
        self.max_atoms = max_atoms

    def get_generator(self) -> GB_MCTS_Generator:
        """Create an instance of the GB_MCTS_Generator.

        Returns:
            an instance of GB_MCTS_Generator.
        """
        optimiser = GB_MCTS_Generator(
            pickle_directory=self.pickle_directory,
            init_smiles=self.init_smiles,
            population_size=self.population_size,
            max_children=self.max_children,
            num_sims=self.num_sims,
            generations=self.generations,
            n_jobs=self.n_jobs,
            max_atoms=self.max_atoms,
            patience=self.patience,
        )
        return optimiser



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""TorchDrug GraphAF training utilities."""
import ast
import logging
import os
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, Optional

from torchdrug.core import Engine
from torchdrug.layers import distribution
from torchdrug.models import RGCN, GraphAF
from torchdrug.tasks import AutoregressiveGeneration

# isort: off
import torch
from torch import optim
from torch import nn

# isort: on
from ....cli.argument_parser import eval_lambda
from ...core import TrainingPipelineArguments
from .. import DATASET_FACTORY
from ..core import TorchDrugTrainingPipeline

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

"""
Necessary because torchdrug silently overwrites the default nn.Module. This is quite
invasive and causes significant side-effects in the rest of the code.
See: https://github.com/DeepGraphLearning/torchdrug/issues/77
"""
nn.Module = nn._Module  # type: ignore


class TorchDrugGraphAFTrainingPipeline(TorchDrugTrainingPipeline):
    """TorchDrug GraphAF training pipelines."""

    def train(  # type: ignore
        self,
        training_args: Dict[str, Any],
        model_args: Dict[str, Any],
        dataset_args: Dict[str, Any],
    ) -> None:
        """Generic training function for training a
            (GraphAF) model. For details see:
                Shi, Chence, et al. "GraphAF: a Flow-based Autoregressive Model for
                Molecular Graph Generation".
                International Conference on Learning Representations (ICLR), 2020.

        Args:
            training_args: training arguments passed to the configuration.
            model_args: model arguments passed to the configuration.
            dataset_args: dataset arguments passed to the configuration.
        """
        try:

            params = {**training_args, **dataset_args, **model_args}

            model_path = params["model_path"]
            training_name = params["training_name"]
            dataset_name = params["dataset_name"]

            logger.info(f"Model with name {training_name} starts.")

            model_dir = Path(model_path).joinpath(training_name)
            os.makedirs(model_dir, exist_ok=True)

            # Set up the dataset here
            joint_dataset_args = {
                "verbose": params.get("verbose", 1),
                "lazy": params.get("lazy", False),
                "transform": eval_lambda(params.get("transform", "lambda x: x")),
                "node_feature": params.get("node_feature", "default"),
                "edge_feature": params.get("edge_feature", "default"),
                "graph_feature": params.get("graph_feature", None),
                "with_hydrogen": params.get("with_hydrogen", False),
                "kekulize": not params.get("no_kekulization", False),
            }
            if dataset_name not in DATASET_FACTORY.keys():
                raise ValueError(
                    f"Dataset {dataset_name} is not supported. Choose from "
                    f"{DATASET_FACTORY.keys()}"
                )
            if dataset_name != "custom":
                # This is a native TorchDrug dataset
                dataset = DATASET_FACTORY[dataset_name](
                    path=params["dataset_path"], **joint_dataset_args
                )
            else:
                # User brought their own dataset
                dataset = DATASET_FACTORY["custom"](
                    file_path=params["file_path"],
                    target_fields=[params["target_field"]],
                    smiles_field=params.get("smiles_field", "smiles"),
                    **joint_dataset_args,
                )

            hidden_dims = ast.literal_eval(params["hidden_dims"])
            num_atom_type = dataset.num_atom_type
            num_bond_type = dataset.num_bond_type

            model = RGCN(
                input_dim=params.get("input_dim", num_atom_type),
                num_relation=params.get("num_relation", num_bond_type),
                hidden_dims=hidden_dims,
                batch_norm=params.get("batch_norm", True),
                edge_input_dim=params.get("edge_input_dim", None),
                short_cut=params.get("short_cut", False),
                activation=params.get("activation", "relu"),
                concat_hidden=params.get("concat_hidden", False),
                readout=params.get("readout", "sum"),
            )
            task = params.get("task")
            if dataset_name == "custom" and task and params["target_field"] != task:
                raise ValueError(
                    "If custom dataset is used & task is specified, then target_field "
                    "has to be set s.t. it extracts the task/property of interest. "
                    f"Not task={task} and target_field={params['target_field']}"
                )
            criterion = ast.literal_eval(params["criterion"])
            if "ppo" in criterion.keys() and (
                params["no_kekulization"] or params["node_feature"] != "symbol"
            ):
                # See torchdrug issue: https://github.com/DeepGraphLearning/torchdrug/issues/77
                raise ValueError(
                    "For property optimiz. leave `no_kekulization` at the default ("
                    "False) & set `node_feature` to `symbol` and not: "
                    f"{params['no_kekulization']} and `{params['node_feature']}`."
                )

            # Model prior/flow initialization
            node_prior = distribution.IndependentGaussian(
                torch.zeros(num_atom_type), torch.ones(num_atom_type)
            )
            edge_prior = distribution.IndependentGaussian(
                torch.zeros(num_bond_type + 1), torch.ones(num_bond_type + 1)
            )
            node_flow = GraphAF(
                model, node_prior, num_layer=params.get("num_node_flow_layers", 12)
            )
            edge_flow = GraphAF(
                model,
                edge_prior,
                num_layer=params.get("num_edge_flow_layers", 12),
                use_edge=not params.get("no_edge", False),
            )

            task = AutoregressiveGeneration(
                node_flow,
                edge_flow,
                task=params.get("task", None),
                max_edge_unroll=params.get("max_edge_unroll", 12),
                max_node=params.get("max_node", 38),
                criterion=ast.literal_eval(params["criterion"]),
                num_node_sample=params.get("num_node_sample", -1),
                num_edge_sample=params.get("num_edge_sample", -1),
                agent_update_interval=params.get("agent_update_interval", 10),
                gamma=params.get("gamma", 0.9),
                reward_temperature=params.get("reward_temperature", 1.0),
                baseline_momentum=params.get("baseline_momentum", 0.9),
            )

            optimizer = optim.Adam(
                task.parameters(), lr=params.get("learning_rate", 1e-5)
            )
            device = (0,) if torch.cuda.is_available() else None
            solver = Engine(
                task,
                dataset,
                None,  # validation data
                None,  # test data
                optimizer,
                batch_size=params.get("batch_size", 16),
                log_interval=params.get("log_interval", 100),
                scheduler=params.get("scheduler", None),
                gpus=params.get("gpus", device),
                gradient_interval=params.get("gradient_interval", 1),
                num_worker=params.get("num_worker", 0),
            )
            # Necessary since we have re-assigned nn.Module to the native torch.nn.Module
            # rather than the torchdrug-overwritten version.
            solver.model.device = solver.device

            weight_paths = sorted(list(model_dir.glob("*.pkl")), key=os.path.getmtime)
            if len(weight_paths) > 0:
                solver.load(
                    weight_paths[-1], load_optimizer=params.get("load_optimizer", False)
                )
                logger.info(f"Restored existing model from {weight_paths[-1]}")
                logger.info(
                    "To avoid this, set `training_name` & `model_path` to a new folder."
                )

            epochs = params.get("epochs", 10)
            solver.train(num_epoch=epochs)

            # Save model
            task_name = f"task={params.get('task')}_" if params.get("task") else ""
            data_name = "data=" + (
                dataset_name
                + "_"
                + str(params["file_path"]).split(os.sep)[-1].split(".")[0]
                if dataset_name == "custom"
                else dataset_name
            )

            solver.save(
                model_dir.joinpath(
                    f"graphaf_data={data_name}_{task_name}epoch={epochs}.pkl"
                )
            )

        except Exception:
            logger.exception(
                "Exception occurred while running TorchDrugGraphAFTrainingPipeline"
            )


@dataclass
class TorchDrugGraphAFModelArguments(TrainingPipelineArguments):
    """Arguments pertaining to model instantiation."""

    __name__ = "model_args"

    hidden_dims: str = field(
        default="[128, 128]",
        metadata={"help": "Dimensionality of each hidden layer"},
    )
    batch_norm: bool = field(
        default=False, metadata={"help": "Whether the RGCN uses batch normalization"}
    )
    edge_input_dim: Optional[int] = field(
        default=None, metadata={"help": "Dimension of edge features"}
    )
    short_cut: bool = field(
        default=False, metadata={"help": "Whether the RGCN uses a short cut"}
    )
    activation: str = field(
        default="relu", metadata={"help": "Activation function for RGCN"}
    )
    concat_hidden: bool = field(
        default=False,
        metadata={
            "help": "Whether hidden representations from all layers are concatenated"
        },
    )
    num_node_flow_layers: int = field(
        default=12,
        metadata={"help": "Number of layers in the node flow GraphAF model"},
    )
    num_edge_flow_layers: int = field(
        default=12,
        metadata={"help": "Number of layers in the edge flow GraphAF model"},
    )
    no_edge: bool = field(
        default=False,
        metadata={
            "help": "Whether to use edge features in the edge GraphAF model. Per "
            "default, edges are used."
        },
    )

    readout: str = field(
        default="sum",
        metadata={"help": "RGCN Readout function. Either `sum` or `mean`"},
    )
    max_edge_unroll: int = field(
        default=12,
        metadata={
            "help": "max node id difference. Inferred from training data if not provided"
        },
    )
    max_node: int = field(
        default=38,
        metadata={
            "help": "max number of node. Inferred from training data if not provided."
        },
    )
    criterion: str = field(
        default="{'nll': 1.0}",
        metadata={
            "help": "training criterion. Available criteria are `nll` and `ppo` for"
            " regular training and property optimization respectively. If dict, the "
            "keys are criterions and values are the corresponding weights. If list, "
            "both criteria are used with equal weights."
        },
    )
    num_node_sample: int = field(
        default=-1,
        metadata={"help": "Number of node samples per graph."},
    )
    num_edge_sample: int = field(
        default=-1,
        metadata={"help": "Number of edge samples per graph."},
    )
    agent_update_interval: int = field(
        default=10,
        metadata={
            "help": "Update the agent every n batches (similar to gradient accumulation)"
        },
    )
    gamma: float = field(
        default=0.9,
        metadata={"help": "Reward discount rate"},
    )
    reward_temperature: float = field(
        default=1.0,
        metadata={
            "help": "Temperature for the reward (larger -> higher mean reward)"
            "lower -> higher maximal reward."
        },
    )
    baseline_momentum: float = field(
        default=0.9,
        metadata={"help": "Momentum for value function baseline"},
    )



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""PaccMannVAE Algorithm.

PaccMannVAE is an unconditional molecular generative model.
"""

import logging
from dataclasses import field
from typing import Any, ClassVar, Dict, Optional, Set, TypeVar

from ....domains.materials import SMILES
from ...conditional_generation.paccmann_rl.core import PaccMannRLProteinBasedGenerator
from ...core import AlgorithmConfiguration, GeneratorAlgorithm, Untargeted
from ...registry import ApplicationsRegistry
from .implementation import PaccMannVaeDefaultGenerator

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

T = type(None)
S = TypeVar("S", bound=SMILES)


class PaccMannVAE(GeneratorAlgorithm[S, T]):
    """Molecular VAE as in the PaccMann\\ :superscript:`RL` paper."""

    def __init__(
        self, configuration: AlgorithmConfiguration[S, T], target: Optional[T] = None
    ):
        """Instantiate PaccMannVAE ready to generate molecules.

        Args:
            configuration: domain and application specification defining parameters,
                types and validations.
            target: unused since it is not a conditional generator.

        Example:
            An example for unconditional generation of small molecules::

                config = PaccMannVAEGenerator()
                algorithm = PaccMannVAE(configuration=config)
                items = list(algorithm.sample(10))
                print(items)
        """

        configuration = self.validate_configuration(configuration)
        super().__init__(configuration=configuration, target=None)  # type:ignore

    def get_generator(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ) -> Untargeted:
        """Get the function to sample batches via PaccMannVAE.

        Args:
            configuration: helps to set up specific application of PaccMannVAE.

        Returns:
            callable with target generating a batch of items.
        """
        implementation: PaccMannVaeDefaultGenerator = (
            configuration.get_conditional_generator()  # type: ignore
        )
        return implementation.generate


@ApplicationsRegistry.register_algorithm_application(PaccMannVAE)
class PaccMannVAEGenerator(AlgorithmConfiguration[SMILES, Any]):
    """
    Configuration to generate molecules with PaccMannVAE.

    Implementation from the paper: https://doi.org/10.1016/j.isci.2021.102269
    """

    algorithm_type: ClassVar[str] = "generation"
    domain: ClassVar[str] = "materials"
    algorithm_version: str = "v0"

    batch_size: int = field(
        default=32,
        metadata=dict(description="Batch size used for the generative model sampling."),
    )
    temperature: float = field(
        default=1.4,
        metadata=dict(
            description="Temperature parameter for the softmax sampling in decoding."
        ),
    )
    generated_length: int = field(
        default=100,
        metadata=dict(
            description="Maximum length in tokens of the generated molcules (relates to the SMILES length)."
        ),
    )

    def get_target_description(self) -> Optional[Dict[str, str]]:
        """Get description of the target for generation.

        Returns:
            target description.
        """
        return None

    def get_conditional_generator(self) -> PaccMannVaeDefaultGenerator:
        """Instantiate the actual generator implementation.

        Args:
            resources_path: local path to model files.

        Returns:
            instance with :meth:`generate_batch<gt4sd.algorithms.conditional_generation.paccmann_rl.implementation.ConditionalGenerator.generate_batch>` method for targeted generation.
        """

        return PaccMannVaeDefaultGenerator(
            temperature=self.temperature,
            generated_length=self.generated_length,
            algorithm_version=self.algorithm_version,
            batch_size=self.batch_size,
        )

    @classmethod
    def list_versions(cls) -> Set[str]:
        """Get possible algorithm versions.

        S3 is searched as well as the local cache is searched for matching versions.

        Returns:
            viable values as :attr:`algorithm_version` for the environment.
        """

        return PaccMannRLProteinBasedGenerator().list_versions()



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Configuration tests."""

import os

from gt4sd.configuration import (
    GT4SDConfiguration,
    get_algorithm_subdirectories_in_cache,
    get_algorithm_subdirectories_with_s3,
)

gt4sd_configuration_instance = GT4SDConfiguration.get_instance()


def test_default_local_cache_path():
    if "GT4SD_LOCAL_CACHE_PATH" not in os.environ:
        assert os.path.dirname(
            gt4sd_configuration_instance.gt4sd_local_cache_path
        ) == os.path.expanduser("~")
        assert (
            os.path.basename(gt4sd_configuration_instance.gt4sd_local_cache_path)
            == ".gt4sd"
        )
    else:
        assert (
            gt4sd_configuration_instance.gt4sd_local_cache_path
            == os.environ["GT4SD_LOCAL_CACHE_PATH"]
        )


def test_get_algorithm_subdirectories_with_s3():
    assert isinstance(get_algorithm_subdirectories_with_s3(), set)


def test_get_algorithm_subdirectories_in_cache():
    assert isinstance(get_algorithm_subdirectories_in_cache(), set)



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import json
import logging
import os
import pickle
from typing import Any, Dict, List, Optional, Tuple

import torch
from rdkit import Chem
from diffusers import DDPMScheduler
from IPython.display import SVG, display
from nglview import show_rdkit as show
from rdkit.Chem.Draw import rdMolDraw2D as MD2
from torch_geometric.data import Data
from torch_scatter import scatter_mean

from .model.core import MoleculeGNN
from .model.utils import repeat_data, set_rdmol_positions

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class GeoDiffPipeline:
    """Pipeline for molecular conformation generation using GeoDiff.
    The pipeline defined here is slightly different than the pipeline used in diffusers.

    GeoDiff: a Geometric Diffusion Model for Molecular Conformation Generation, Minkai Xu, Lantao Yu, Yang Song, Chence Shi, Stefano Ermon, Jian Tang - https://arxiv.org/abs/2203.02923
    """

    def __init__(
        self, model_name_or_path: str, params_json: Optional[str] = None
    ) -> None:
        """GeoDiff pipeline for molecular conformation generation. Code adapted from colab:
                https://colab.research.google.com/drive/1pLYYWQhdLuv1q-JtEHGZybxp2RBF8gPs#scrollTo=-3-P4w5sXkRU written by Nathan Lambert.

        Args:
            model_name_or_path: pretrained model name or path to model directory.
            params_json: parameters as a JSON file. Defaults to None, a.k.a., use default configuration.
        """

        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.model_name_or_path = model_name_or_path

        config = {}
        if params_json is not None:
            with open(params_json, "r") as f:
                config = json.load(f)

        self.scheduler = DDPMScheduler(
            num_train_timesteps=getattr(config, "num_timesteps", 1000),
            beta_schedule=getattr(config, "beta_schedule", "linear"),
            beta_start=getattr(config, "beta_start", 1e-7),
            beta_end=getattr(config, "beta_end", 2e-3),
            clip_sample=getattr(config, "clip_sample", False),
        )
        sigmas = (
            torch.tensor(1.0 - self.scheduler.alphas_cumprod).sqrt()
            / torch.tensor(self.scheduler.alphas_cumprod).sqrt()
        )
        self.sigmas = sigmas.to(self.device)

        self.num_samples = 1  # solutions per molecule
        self.num_molecules = 3
        self.w_global = 0.5  # 0,.3 for qm9
        self.global_start_sigma = 0.5
        self.eta = 1.0
        self.clip_local = None
        self.clip_pos = None

        # constants for data handling
        self.save_traj = False
        self.save_data = False
        self.output_dir = "out"
        os.makedirs(self.output_dir, exist_ok=True)

        # load pretrained model
        self.model = MoleculeGNN.from_pretrained(model_name_or_path)

    def to(self, device: str = "cuda") -> None:
        """Move model to a device.

        Args:
            device: device where to move the model. Defaults to "cuda".
        """
        self.model.to(self.device)

    @classmethod
    def from_pretrained(
        self, model_name_or_path: str, params_json: Optional[str] = None
    ) -> "GeoDiffPipeline":
        """Load pretrained model.

        Args:
            model_name_or_path: pretrained model name or path to model directory.
            params_json: path to model config.

        Returns:
            a GeoDiff pipeline.
        """
        return GeoDiffPipeline(
            model_name_or_path=model_name_or_path, params_json=params_json
        )

    @torch.no_grad()
    def __call__(
        self, batch_size: int, prompt: Dict[str, Any]
    ) -> Dict[str, List[Chem.Mol]]:
        """Generate conformations for a molecule.

        Args:
            batch_size: number of samples to generate.
            prompt: `torch_geometric.data.Data` object containing the molecular graph in 2D format. This information is given as conditioning for the model.

        Returns:
            a dict containing a list of postprocessed generated conformations.
        """

        results = []
        # 2d representation for the molecule
        # convert dict in torch_geometric.data.Data
        data = Data.from_dict(prompt)
        num_samples = max(data.pos_ref.size(0) // data.num_nodes, 1)

        data_input = data.clone()
        data_input["pos_ref"] = None
        batch = repeat_data(data_input, num_samples).to(self.device)

        # initial configuration
        pos_init = torch.randn(batch.num_nodes, 3).to(self.device)

        # for logging animation of denoising
        pos_traj = []

        # sample from pretrained diffusion process
        with torch.no_grad():
            # scale initial sample
            pos = pos_init * self.sigmas[-1]
            for t in self.scheduler.timesteps:
                batch.pos = pos

                # generate geometry with model, then filter it
                epsilon = self.model.forward(
                    batch, t, sigma=self.sigmas[t], return_dict=False
                )[0]

                # Update
                reconstructed_pos = self.scheduler.step(epsilon, t, pos)[
                    "prev_sample"
                ].to(self.device)

                pos = reconstructed_pos

                if torch.isnan(pos).any():
                    raise FloatingPointError("NaN detected. Please restart.")

                # recenter graph of positions for next iteration
                pos = pos - scatter_mean(pos, batch.batch, dim=0)[batch.batch]

                # optional clipping
                if self.clip_pos is not None:
                    pos = torch.clamp(pos, min=-self.clip_pos, max=self.clip_pos)
                pos_traj.append(pos.clone().cpu())

        pos_gen = pos.cpu()

        if self.save_traj:
            pos_gen_traj = [pt.cpu() for pt in pos_traj]
            data.pos_gen = torch.stack(pos_gen_traj)
        else:
            data.pos_gen = pos_gen

        results.append(data)

        if self.save_data:
            save_path = os.path.join(self.output_dir, "samples.pkl")

            with open(save_path, "wb") as f:
                pickle.dump(results, f)

        mols_gen, mols_orig = self.postprocess_output(results)
        return {"sample": mols_gen}

    def postprocess_output(
        self, results: List[Data]
    ) -> Tuple[List[Chem.Mol], List[Chem.Mol]]:
        """Postprocess output of diffusion pipeline.

        Args:
            results: list of `torch_geometric.data.Data` objects containing the molecular graph in 3D format.

        Returns:
            tuple with list of postprocessed generated conformations and list of postprocessed original conformations.
        """

        # the model can generate multiple conformations per 2d geometry
        # num_gen = results[0]["pos_gen"].shape[0]
        # init storage objects
        mols_gen = []
        mols_orig = []
        for to_process in results:

            # store the reference 3d position
            to_process["pos_ref"] = to_process["pos_ref"].reshape(
                -1, to_process["rdmol"].GetNumAtoms(), 3
            )

            # store the generated 3d position
            to_process["pos_gen"] = to_process["pos_gen"].reshape(
                -1, to_process["rdmol"].GetNumAtoms(), 3
            )

            # copy data to new object
            new_mol = set_rdmol_positions(to_process.rdmol, to_process["pos_gen"][0])

            # append results
            mols_gen.append(new_mol)
            mols_orig.append(to_process.rdmol)

        logger.info(f"collect {len(mols_gen)} generated molecules in `mols`")
        return mols_gen, mols_orig

    def visualize_2d_input(self, data: Data) -> None:
        """Visualize 2D input.

        Args:
            data: `torch_geometric.data.Data` object containing the molecular graph in 2D format.
        """

        mc = Chem.MolFromSmiles(data[0]["smiles"])
        molSize = (450, 300)
        drawer = MD2.MolDraw2DSVG(molSize[0], molSize[1])
        drawer.DrawMolecule(mc)
        drawer.FinishDrawing()
        svg = drawer.GetDrawingText()
        display(SVG(svg.replace("svg:", "")))

    def visualize_3d(self, mols_gen: List[Chem.Mol]) -> None:
        """Visualize 3D output.

        Args:
            mols_gen: list of generated conformations.
        """
        show(mols_gen[0])



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test for properties."""
import importlib_resources
import numpy as np
import pytest

from gt4sd.properties import PROPERTY_PREDICTOR_FACTORY, PropertyPredictorRegistry
from gt4sd.properties.crystals import CRYSTALS_PROPERTY_PREDICTOR_FACTORY
from gt4sd.properties.molecules import MOLECULE_PROPERTY_PREDICTOR_FACTORY
from gt4sd.properties.molecules.core import SimilaritySeed
from gt4sd.properties.proteins import PROTEIN_PROPERTY_PREDICTOR_FACTORY
from gt4sd.properties.proteins.core import Charge

protein = "KFLIYQMECSTMIFGL"
molecule = "C1=CC(=CC(=C1)Br)CN"
seed = "CCO"
target = "drd2"
ground_truths = {
    "length": 16,
    "protein_weight": 1924.36,
    "boman_index": -0.534375,
    "aliphaticity": 97.5,
    "hydrophobicity": 0.5625,
    "charge": -0.071,
    "charge_density": -3.6895383400195386e-05,
    "isoelectric_point": 6.125,
    "aromaticity": 0.1875,
    "instability": 36.8375,
    "plogp": 0.25130060815905964,
    "molecular_weight": 186.05199999999996,
    "lipinski": 1,
    "esol": -2.6649954522215555,
    "scscore": 1.6081393182467014,
    "sas": 1.6564993918409403,
    "bertz": 197.86256853719752,
    "tpsa": 26.02,
    "logp": 1.9078,
    "qed": 0.7116306772337652,
    "number_of_h_acceptors": 1,
    "number_of_atoms": 17,
    "number_of_h_donors": 1,
    "number_of_aromatic_rings": 1,
    "number_of_rings": 1,
    "number_of_large_rings": 0,
    "number_of_rotatable_bonds": 1,
    "is_scaffold": 0,
    "number_of_heterocycles": 0,
    "number_of_stereocenters": 0,
}

molecule_further_ground_truths = {
    "activity_against_target": 0.0049787821116125345,
    "similarity_seed": 0.03333333333333333,
}
protein_further_ground_truths = {"charge": 1.123}

artifact_model_data = {
    "tox21": {
        "parameters": {"algorithm_version": "v0"},
        "ground_truth": [
            6.422001024475321e-05,
            0.00028556393226608634,
            0.0027144483756273985,
            0.03775344416499138,
            0.000604992441367358,
            0.00027705798856914043,
            0.10752066224813461,
            0.5733309388160706,
            0.001268531079404056,
            0.10181115567684174,
            0.8995946049690247,
            0.1677667647600174,
        ],
    },
    "organtox": {
        "parameters": {
            "algorithm_version": "v0",
            "site": "Heart",
            "toxicity_type": "all",
        },
        "ground_truth": [0.06142323836684227, 0.07934761792421341],
    },
}

crystal_ground_truths = {
    "formation_energy": -2.1132538318634033,
    "absolute_energy": -3.3883044719696045,
    "band_gap": 4.774056434631348,
    "fermi_energy": -0.6838343143463135,
    "bulk_moduli": 1.3808047771453857,
    "shear_moduli": 0.9817742705345154,
    "poisson_ratio": 0.3124755620956421,
    "metal_semiconductor_classifier": 4.423607151693432e-06,
}


@pytest.mark.parametrize("property_key", crystal_ground_truths.keys())
def test_crystals(property_key: str):

    property_class, parameters_class = CRYSTALS_PROPERTY_PREDICTOR_FACTORY[property_key]
    model = property_class(parameters_class(algorithm_version="v0"))  # type: ignore

    with importlib_resources.as_file(
        importlib_resources.files("gt4sd") / "properties/tests/",
    ) as file_path:
        out = model(input=file_path)  # type: ignore
        pred_dict = dict(zip(out["cif_ids"], out["predictions"]))  # type: ignore
        prediction = pred_dict["1000041"]
        assert np.isclose(prediction, crystal_ground_truths[property_key], atol=1e-2)


def test_rfc():
    property_class, parameters_class = CRYSTALS_PROPERTY_PREDICTOR_FACTORY[
        "metal_nonmetal_classifier"
    ]
    model = property_class(parameters_class(algorithm_version="v0"))  # type: ignore

    with importlib_resources.as_file(
        importlib_resources.files("gt4sd") / "properties/tests/crf_data.csv",
    ) as file_path:

        out = model(input=file_path)  # type: ignore

        pred_dict = dict(zip(out["formulas"], out["predictions"]))  # type: ignore

        assert pred_dict["AgHgHW6"] == "metal"
        assert pred_dict["AlGaH6Os"] == "non-metal"
        assert pred_dict["BaAlTlH6"] == "non-metal"
        assert pred_dict["BaCaH6Ir"] == "non-metal"
        assert pred_dict["BaCaH6Rh"] == "non-metal"
        assert pred_dict["BaCaHfSi6"] == "metal"
        assert pred_dict["BaCrH6Ru"] == "non-metal"
        assert pred_dict["BaMgZnH6"] == "non-metal"
        assert pred_dict["BaMnVH6"] == "non-metal"


@pytest.mark.parametrize(
    "property_key", [(property_key) for property_key in ground_truths.keys()]
)
def test_property(property_key):
    property_class, parameters_class = PROPERTY_PREDICTOR_FACTORY[property_key]
    function = property_class(parameters_class())
    sample = protein if property_key in PROTEIN_PROPERTY_PREDICTOR_FACTORY else molecule
    assert np.isclose(function(sample), ground_truths[property_key], atol=1e-2)  # type: ignore


def test_similarity_seed():
    property_class, parameters_class = MOLECULE_PROPERTY_PREDICTOR_FACTORY[
        "similarity_seed"
    ]
    function = property_class(parameters_class(smiles=seed))  # type: ignore
    assert np.isclose(
        function(molecule), molecule_further_ground_truths["similarity_seed"], atol=1e-2  # type: ignore
    )


def test_activity_against_target():
    property_class, parameters_class = MOLECULE_PROPERTY_PREDICTOR_FACTORY[
        "activity_against_target"
    ]
    function = property_class(parameters_class(target=target))  # type: ignore
    assert np.isclose(
        function(molecule), molecule_further_ground_truths["activity_against_target"], atol=1e-2  # type: ignore
    )


def test_charge_with_arguments():
    property_class, parameters_class = PROTEIN_PROPERTY_PREDICTOR_FACTORY["charge"]
    function = property_class(parameters_class(amide=True, ph=5.0))
    assert np.isclose(function(protein), protein_further_ground_truths["charge"], atol=1e-2)  # type: ignore


@pytest.mark.parametrize(
    "property_key",
    [(property_key) for property_key in artifact_model_data.keys()],
)
def test_artifact_models(property_key):
    property_class, parameters_class = PROPERTY_PREDICTOR_FACTORY[property_key]
    function = property_class(
        parameters_class(**artifact_model_data[property_key]["parameters"])
    )
    sample = protein if property_key in PROTEIN_PROPERTY_PREDICTOR_FACTORY else molecule
    assert all(
        np.isclose(
            function(sample),
            artifact_model_data[property_key]["ground_truth"],  # type: ignore
            atol=1e-2,
        )
    )  # type: ignore


def test_property_predictor_registry():
    predictor = PropertyPredictorRegistry.get_property_predictor(
        "similarity_seed", {"smiles": seed}
    )
    assert isinstance(predictor, SimilaritySeed)
    assert np.isclose(
        predictor(molecule), molecule_further_ground_truths["similarity_seed"]  # type: ignore
    )
    predictor = PropertyPredictorRegistry.get_property_predictor(
        "charge", {"amide": "True", "ph": 5.0}
    )
    assert isinstance(predictor, Charge)
    assert np.isclose(predictor(protein), protein_further_ground_truths["charge"])  # type: ignore
    assert len(PropertyPredictorRegistry.list_available()) == len(
        PROTEIN_PROPERTY_PREDICTOR_FACTORY
    ) + len(MOLECULE_PROPERTY_PREDICTOR_FACTORY) + len(
        CRYSTALS_PROPERTY_PREDICTOR_FACTORY
    )



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from typing import Any, Dict, Tuple

import torch
import torch.nn as nn
import torch_geometric.data as gd
import torch_geometric.nn as gnn
from torch_geometric.utils import add_self_loops

from ...envs.graph_building_env import GraphActionCategorical, GraphBuildingEnvContext


def mlp(n_in, n_hid, n_out, n_layer, act=nn.LeakyReLU):
    n = [n_in] + [n_hid] * n_layer + [n_out]
    return nn.Sequential(
        *sum([[nn.Linear(n[i], n[i + 1]), act()] for i in range(n_layer + 1)], [])[:-1]
    )


class GraphTransformer(nn.Module):
    """GraphTransformer."""

    def __init__(
        self,
        configuration: Dict[str, Any],
        context: GraphBuildingEnvContext,
        x_dim=64,
        e_dim=64,
        g_dim=64,
        num_emb=64,
        num_layers=3,
        num_heads=2,
    ) -> None:
        """Construct GraphTransformer.

        Code adapted from: https://github.com/recursionpharma/gflownet/tree/trunk/src/gflownet/models.

        Args:
            configuration: model configuration.
            context: context environment.
            x_dim: dimension of input node features.
            e_dim: dimension of input edge features.
            g_dim: dimension of input graph features.
            num_emb: dimension of embedding layer.
            num_layers: number of layers in the graph transformer.
            num_heads: number of heads in the graph transformer.
        """

        super().__init__()
        self.num_layers = num_layers
        self.name = "graph_transformer"

        self.x2h = mlp(x_dim, num_emb, num_emb, 2)
        self.e2h = mlp(e_dim, num_emb, num_emb, 2)
        self.c2h = mlp(g_dim, num_emb, num_emb, 2)
        self.graph2emb = nn.ModuleList(
            sum(
                [
                    [
                        gnn.GENConv(
                            num_emb, num_emb, num_layers=1, aggr="add", norm=None
                        ),
                        gnn.TransformerConv(
                            num_emb * 2, num_emb, edge_dim=num_emb, heads=num_heads
                        ),
                        nn.Linear(num_heads * num_emb, num_emb),
                        gnn.LayerNorm(num_emb, affine=False),
                        mlp(num_emb, num_emb * 4, num_emb, 1),
                        gnn.LayerNorm(num_emb, affine=False),
                    ]
                    for i in range(self.num_layers)
                ],
                [],
            )
        )

    def forward(
        self, g: gd.Batch, cond: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Forward pass.

        Args:
            g: graph data.
            cond: conditioning.

        Returns:
            embeddings and pooled features.
        """
        o = self.x2h(g.x)
        e = self.e2h(g.edge_attr)
        c = self.c2h(cond)
        num_total_nodes = g.x.shape[0]
        # Augment the edges with a new edge to the conditioning
        # information node. This new node is connected to every node
        # within its graph.
        u, v = torch.arange(num_total_nodes, device=o.device), g.batch + num_total_nodes
        aug_edge_index = torch.cat(
            [g.edge_index, torch.stack([u, v]), torch.stack([v, u])], 1
        )
        e_p = torch.zeros((num_total_nodes * 2, e.shape[1]), device=g.x.device)
        e_p[:, 0] = 1  # Manually create a bias term
        aug_e = torch.cat([e, e_p], 0)
        aug_edge_index, aug_e = add_self_loops(aug_edge_index, aug_e, "mean")
        aug_batch = torch.cat([g.batch, torch.arange(c.shape[0], device=o.device)], 0)

        # Append the conditioning information node embedding to o
        o = torch.cat([o, c], 0)
        for i in range(self.num_layers):
            # Run the graph transformer forward
            gen, trans, linear, norm1, ff, norm2 = self.graph2emb[i * 6 : (i + 1) * 6]
            agg = gen(o, aug_edge_index, aug_e)
            o = norm1(
                o + linear(trans(torch.cat([o, agg], 1), aug_edge_index, aug_e)),
                aug_batch,
            )
            o = norm2(o + ff(o), aug_batch)

        glob = torch.cat(
            [gnn.global_mean_pool(o[: -c.shape[0]], g.batch), o[-c.shape[0] :], c], 1
        )
        o_final = torch.cat([o[: -c.shape[0]], c[g.batch]], 1)
        return o_final, glob


class GraphTransformerGFN(nn.Module):
    """GraphTransformerGFN."""

    def __init__(
        self,
        configuration: Dict[str, Any],
        context: GraphBuildingEnvContext,
        num_emb=64,
        num_layers=3,
        num_heads=2,
    ) -> None:
        """Construct GraphTransformerGFN.

        Args:
            configuration: model configuration.
            context: context environment.
            num_emb: dimension of embedding layer.
            num_layers: number of layers in the graph transformer.
            num_heads: number of heads in the graph transformer.
        """

        super().__init__()
        self.name = "graph_transformer_gfn"
        self.transf = GraphTransformer(
            configuration=configuration,
            context=context,
            x_dim=context.num_node_dim,
            e_dim=context.num_edge_dim,
            g_dim=context.num_cond_dim,
            num_emb=num_emb,
            num_layers=num_layers,
            num_heads=num_heads,
        )
        num_final = num_emb * 2
        num_mlp_layers = 0
        self.emb2add_edge = mlp(num_final, num_emb, 1, num_mlp_layers)
        self.emb2add_node = mlp(
            num_final, num_emb, context.num_new_node_values, num_mlp_layers
        )
        self.emb2set_node_attr = mlp(
            num_final, num_emb, context.num_node_attr_logits, num_mlp_layers
        )
        self.emb2set_edge_attr = mlp(
            num_final, num_emb, context.num_edge_attr_logits, num_mlp_layers
        )
        self.emb2stop = mlp(num_emb * 3, num_emb, 1, num_mlp_layers)
        self.emb2reward = mlp(num_emb * 3, num_emb, 1, num_mlp_layers)
        self.log_z = mlp(context.num_cond_dim, num_emb * 2, 1, 2)
        self.action_type_order = context.action_type_order

    def forward(self, g: gd.Batch, cond: torch.Tensor):
        """Forward pass. Given a graph and a conditioning, return the action logits and rewards.

        Args:
            g: graph data.
            cond: conditioning.

        Returns:
            categorical action and rewards.
        """
        node_embeddings, graph_embeddings = self.transf(g, cond)
        ne_row, ne_col = g.non_edge_index
        # On `::2`, edges are duplicated to make graphs undirected, only take the even ones
        e_row, e_col = g.edge_index[:, ::2]
        cat = GraphActionCategorical(
            g,
            logits=[
                self.emb2stop(graph_embeddings),
                self.emb2add_node(node_embeddings),
                self.emb2set_node_attr(node_embeddings),
                self.emb2add_edge(node_embeddings[ne_row] + node_embeddings[ne_col]),
                self.emb2set_edge_attr(node_embeddings[e_row] + node_embeddings[e_col]),
            ],
            keys=[None, "x", "x", "non_edge_index", "edge_index"],
            types=self.action_type_order,
        )
        return cat, self.emb2reward(graph_embeddings)



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""RegressionTransformer algorithm.

RegressionTransformer is a mutlitask regression and conditional generation model.
"""

import logging
import os
from dataclasses import field
from typing import Any, Callable, ClassVar, Dict, Iterable, Optional, TypeVar, Union

from typing_extensions import Protocol, runtime_checkable

from ....domains.materials import Molecule, Sequence
from ....exceptions import InvalidItem
from ....properties.core import PropertyValue
from ....training_pipelines.core import TrainingPipelineArguments
from ....training_pipelines.regression_transformer.core import (
    RegressionTransformerSavingArguments,
)
from ...core import AlgorithmConfiguration, GeneratorAlgorithm
from ...registry import ApplicationsRegistry
from .implementation import ChemicalLanguageRT, ConditionalGenerator, ProteinLanguageRT

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

T = TypeVar("T", bound=Sequence)
S = TypeVar("S", PropertyValue, Molecule)
Targeted = Callable[[T], Iterable[Any]]


class RegressionTransformer(GeneratorAlgorithm[S, T]):
    """RegressionTransformer Algorithm."""

    #: The maximum number of samples a user can try to run in one go
    max_samples: int = 50

    def __init__(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ) -> None:
        """Instantiate Regression Transformer ready to generate items.

        Args:
            configuration: domain and application
                specification defining parameters, types and validations.
            target: a target for which to generate items.

        Example:
            An example for generating small molecules (SMILES) with high affinity for a target protein::

                config = RegressionTransformerProteins(
                    search='sample', temperature=2.0, tolerance=10
                )
                target = "<stab>0.393|GSQEVNSGT[MASK][MASK][MASK]YKNASPEEAE[MASK][MASK]IARKAGATTWTEKGNKWEIRI"
                stability_generator = RegressionTransformer(configuration=config, target=target)
                items = list(stability_generator.sample(10))
                print(items)
        """

        configuration = self.validate_configuration(configuration)

        # No validation/check on the target input here, since model is not yet loaded.
        super().__init__(
            configuration=configuration,  # type:ignore
            target=target,  # type:ignore
        )

    def get_generator(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ) -> Targeted[T]:
        """Get the function to sample with the given configuration.

        Args:
            configuration: helps to set up specific application of PaccMannRL.
            target: context or condition for the generation.

        Returns:
            callable with target generating a batch of items.
        """
        logger.info("ensure artifacts for the application are present.")
        self.local_artifacts = configuration.ensure_artifacts()
        implementation: ConditionalGenerator = configuration.get_conditional_generator(  # type: ignore
            resources_path=self.local_artifacts, context=target
        )
        if implementation.task == "regression" and configuration.search == "greedy":  # type: ignore
            self.max_samples = 1
            logger.warning(
                "max_samples was set to 1 due to regression task and greedy search"
            )

        return implementation.generate_batch  # type: ignore

    def validate_configuration(
        self, configuration: AlgorithmConfiguration[S, T]
    ) -> AlgorithmConfiguration[S, T]:
        @runtime_checkable
        class AnyRegressionTransformerConfiguration(Protocol):
            """Protocol for RegressionTransformer configurations."""

            def get_conditional_generator(
                self, resources_path: str
            ) -> ConditionalGenerator:
                ...

            def validate_item(self, item: Any) -> S:
                ...

        # TODO raise InvalidAlgorithmConfiguration
        assert isinstance(configuration, AnyRegressionTransformerConfiguration)
        assert isinstance(configuration, AlgorithmConfiguration)
        return configuration


@ApplicationsRegistry.register_algorithm_application(RegressionTransformer)
class RegressionTransformerMolecules(AlgorithmConfiguration[Sequence, Sequence]):
    """
    Configuration to generate molecules given a continuous property target and a molecular sub-structure.

    Implementation from the paper: https://arxiv.org/abs/2202.01338.

    Examples:
        An example for generating a peptide around a desired property value::

            config = RegressionTransformerMolecules(
                algorithm_version='solubility', search='sample', temperature=2, tolerance=5
            )
            target = "<esol>-3.534|[Br][C][=C][C][MASK][MASK][=C][C][=C][C][=C][Ring1][MASK][MASK][Branch2_3][Ring1][Branch1_2]"
            solubility_generator = RegressionTransformer(
                configuration=config, target=target
            )
            list(solubility_generator.sample(5))

        An example for predicting the solubility of a molecule::

            config = RegressionTransformerMolecules(
                algorithm_version='solubility', search='greedy'
            )
            target = "<esol>[MASK][MASK][MASK][MASK][MASK]|[Cl][C][Branch1_2][Branch1_2][=C][Branch1_1][C][Cl][Cl][Cl]"
            solubility_generator = RegressionTransformer(
                configuration=config, target=target
            )
            list(solubility_generator.sample(1))
    """

    algorithm_type: ClassVar[str] = "conditional_generation"
    domain: ClassVar[str] = "materials"
    algorithm_version: str = field(
        default="solubility",
        metadata=dict(
            description="The version of the algorithm to use.",
            options=["solubility", "qed", "logp_and_synthesizability"],
        ),
    )

    search: str = field(
        default="sample",
        metadata=dict(
            description="Search algorithm to use for the generation: sample or greedy",
            options=["sample", "greedy"],
        ),
    )

    temperature: float = field(
        default=1.4,
        metadata=dict(
            description="Temperature parameter for the softmax sampling in decoding."
        ),
    )
    batch_size: int = field(
        default=8,
        metadata=dict(description="Batch size for the conditional generation"),
    )
    tolerance: Union[float, Dict[str, float]] = field(
        default=20.0,
        metadata=dict(
            description="""Precision tolerance for the conditional generation task. This is the
            tolerated eviation between desired/primed property and predicted property of the
            generated molecule. Given in percentage with respect to the property range encountered
            during training. Either a single float or a dict of floats with properties as
            NOTE: The tolerance is *only* used for post-hoc filtering of the generated molecules.
            """
        ),
    )
    sampling_wrapper: Dict = field(
        default_factory=dict,
        metadata=dict(
            description="""High-level entry point for SMILES-level access. Provide a
            dictionary that is used to build a custom sampling wrapper.
            NOTE: If this is used, the `target` needs to be a single SMILES string.
            Example: {
                'fraction_to_mask': 0.5,
                'tokens_to_mask': [],
                'property_goal': {'<qed>': 0.85}
            }
            - 'fraction_to_mask' specifies the ratio of tokens that can be changed by
                the model.
            - 'tokens_to_mask' specifies which atoms can be masked. This defaults
                to an empty list, meaning that all tokens can be masked.
            - 'property_goal' specifies the target conditions for the generation. The
                properties need to be specified as a dictionary. The keys need to be
                properties supported by the algorithm version.
            - 'substructures_to_mask': Specifies a list of substructures that should be masked.
                Given in SMILES format. This is excluded from the stochastic masking.
                NOTE: The model operates on SELFIES and the matching of the substructures occurs
                in SELFIES simply on a string level.
            - 'substructures_to_keep': Specifies a list of substructures that should definitely be kept.
                Given in SMILES format. This is excluded from the stochastic masking.
                NOTE: This keeps tokens even if they are included in `tokens_to_mask`.
                NOTE: The model operates on SELFIES and the matching of the substructures occurs
                in SELFIES simply on a string level.
            - `text_filtering`: Generated sequences are post-hoc filtered for the presence of
                `substructures_to_keep`. This is done with RDKit substructure matches. If the sub-
                structure cant be converted to a mol object, this argument toggles whether a substructure
                should be ignored from post-hoc filtering (this happens per default) or whether
                filtering should occur on a pure string level. Defaults to False.
                NOTE: This does not affect the actual generation process.
            """
        ),
    )

    def get_target_description(self) -> Dict[str, str]:
        """Get description of the target for generation.

        Returns:
            target description.
        """
        return {
            "title": "Masked input sequence",
            "description": (
                "A sequence with a property value and a SELFIES string. Masking can either occur on the property or on the SELFIES, but not both."
                "For the scale of the property values, please see the task/dataset."
            ),
            "type": "string",
        }

    def get_conditional_generator(
        self, resources_path: str, context: str
    ) -> ChemicalLanguageRT:
        """Instantiate the actual generator implementation.

        Args:
            resources_path: local path to model files.

        Returns:
            instance with :meth:`generate_batch<gt4sd.algorithms.conditional_generation.regression_transformer.implementation.ChemicalLanguageRT.generate_batch>` method for targeted generation.
        """
        self.generator = ChemicalLanguageRT(
            resources_path=resources_path,
            context=context,
            search=self.search,
            temperature=self.temperature,
            batch_size=self.batch_size,
            tolerance=self.tolerance,
            sampling_wrapper=self.sampling_wrapper,
        )
        return self.generator

    def validate_item(self, item: str) -> Union[Molecule, Sequence]:  # type: ignore
        """Check that item is a valid sequence.

        Args:
            item: a generated item that is possibly not valid.

        Raises:
            InvalidItem: in case the item can not be validated.

        Returns:
            the validated item.
        """
        if item is None:
            raise InvalidItem(title="InvalidSequence", detail="Sequence is None")
        items, _ = self.generator.validate_output([item])
        if items[0] is None:
            if self.generator.task == "generation":
                title = "InvalidSMILES"
                detail = f'rdkit.Chem.MolFromSmiles returned None for "{item}"'
            else:
                title = "InvalidNumerical"
                detail = f'"{item}" is not a valid floating point number'
            raise InvalidItem(title=title, detail=detail)
        return item

    @classmethod
    def get_filepath_mappings_for_training_pipeline_arguments(
        cls, training_pipeline_arguments: TrainingPipelineArguments
    ) -> Dict[str, str]:
        """Get filepath mappings for the given training pipeline arguments.
        Args:
            training_pipeline_arguments: training pipeline arguments.
        Returns:
            a mapping between artifacts' files and training pipeline's output files.
        """
        if isinstance(
            training_pipeline_arguments, RegressionTransformerSavingArguments
        ):
            training_path = os.path.abspath(training_pipeline_arguments.model_path)
            if training_pipeline_arguments.checkpoint_name:
                model_path = os.path.join(
                    training_path, training_pipeline_arguments.checkpoint_name
                )
            else:
                model_path = training_path

            names = ["pytorch_model.bin", "config.json", "vocab.txt"]
            names.extend(["special_tokens_map.json", "tokenizer_config.json"])

            mapper = {name: os.path.join(model_path, name) for name in names}

            # Inference file is only saved once in root training folder
            mapper["inference.json"] = os.path.join(training_path, "inference.json")
            return mapper

        else:
            return super().get_filepath_mappings_for_training_pipeline_arguments(
                training_pipeline_arguments
            )


@ApplicationsRegistry.register_algorithm_application(RegressionTransformer)
class RegressionTransformerProteins(AlgorithmConfiguration[Sequence, Sequence]):
    """
    Configuration to generate protein given a continuous property target and a partial AAs.

    Implementation from the paper: https://arxiv.org/abs/2202.01338. It can also predict the property given a full sequence.

    Examples:
        An example for generating a peptide around a desired property value::

            config = RegressionTransformerProteins(
                search='sample', temperature=2, tolerance=5
            )
            target = "<stab>1.1234|TTIKNG[MASK][MASK][MASK]YTVPLSPEQAAK[MASK][MASK][MASK]KKRWPDYEVQIHGNTVKVT"
            stability_generator = RegressionTransformer(
                configuration=config, target=target
            )
            list(stability_generator.sample(5))

        An example for predicting the stability of a peptide::

            config = RegressionTransformerProteins(search='greedy')
            target = "<stab>[MASK][MASK][MASK][MASK][MASK]|GSQEVNSNASPEEAEIARKAGATTWTEKGNKWEIRI"
            stability_generator = RegressionTransformer(
                configuration=config, target=target
            )
            list(stability_generator.sample(1))
    """

    algorithm_type: ClassVar[str] = "conditional_generation"
    domain: ClassVar[str] = "materials"
    algorithm_version: str = field(
        default="stability",
        metadata=dict(
            description="The version of the algorithm to use.",
            options=["stability"],
        ),
    )

    search: str = field(
        default="sample",
        metadata=dict(
            description="Search algorithm to use for the generation: sample or greedy"
        ),
    )

    temperature: float = field(
        default=1.4,
        metadata=dict(
            description="Temperature parameter for the softmax sampling in decoding."
        ),
    )
    batch_size: int = field(
        default=32,
        metadata=dict(description="Batch size for the conditional generation"),
    )
    tolerance: Union[float, Dict[str, float]] = field(
        default=20.0,
        metadata=dict(
            description="""Precision tolerance for the conditional generation task. This is the
            tolerated eviation between desired/primed property and predicted property of the
            generated molecule. Given in percentage with respect to the property range encountered
            during training. Either a single float or a dict of floats with properties as
            NOTE: The tolerance is *only* used for post-hoc filtering of the generated proteins.
            """
        ),
    )
    sampling_wrapper: Dict = field(
        default_factory=dict,
        metadata=dict(
            description="""High-level entry point for SMILES-level access. Provide a
            dictionary that is used to build a custom sampling wrapper.
            NOTE: If this is used, the `target` needs to be a single SMILES string.
            Example: {
                'fraction_to_mask': 0.5,
                'tokens_to_mask': [],
                'property_goal': {'<qed>': 0.85}
            }
            - 'fraction_to_mask' specifies the ratio of tokens that can be changed by
                the model.
            - 'tokens_to_mask' specifies which atoms can be masked. This defaults
                to an empty list, meaning that all tokens can be masked.
            - 'property_goal' specifies the target conditions for the generation. The
                properties need to be specified as a dictionary. The keys need to be
                properties supported by the algorithm version.
            """
        ),
    )

    def get_target_description(self) -> Dict[str, str]:
        """Get description of the target for generation.

        Returns:
            target description.
        """
        return {
            "title": "Masked input sequence",
            "description": "A sequence with a property value and an AAS. Masking can either occur on the property or on the AAS, but not both.",
            "type": "string",
        }

    def get_conditional_generator(
        self, resources_path: str, context: str
    ) -> ProteinLanguageRT:
        """Instantiate the actual generator implementation.

        Args:
            resources_path: local path to model files.
            context: input sequence to be used for the generation.

        Returns:
            instance with :meth:`generate_batch<gt4sd.algorithms.conditional_generation.regression_transformer.implementation.ProteinLanguageRT.generate_batch>` method for targeted generation.
        """

        self.generator = ProteinLanguageRT(
            resources_path=resources_path,
            search=self.search,
            temperature=self.temperature,
            context=context,
            batch_size=self.batch_size,
            tolerance=self.tolerance,
            sampling_wrapper=self.sampling_wrapper,
        )
        return self.generator

    def validate_item(self, item: str) -> Union[Molecule, Sequence]:  # type: ignore
        """Check that item is a valid sequence.

        Args:
            item: a generated item that is possibly not valid.

        Raises:
            InvalidItem: in case the item can not be validated.

        Returns:
            the validated item.
        """
        if item is None:
            raise InvalidItem(title="InvalidSequence", detail="Sequence is None")
        items, _ = self.generator.validate_output([item])
        if items[0] is None:
            if self.generator.task == "generation":
                title = "InvalidSequence"
                detail = f'"{item}" does not adhere to IUPAC convention for AAS'
            else:
                title = "InvalidNumerical"
                detail = (
                    f'"{item}" is not a valid Sequence with a floating point number'
                )
            raise InvalidItem(title=title, detail=detail)
        return item



#!/usr/bin/env python
#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Run inference pipelines for the GT4SD."""

import json
import logging
import os
import sys
from dataclasses import dataclass, field
from typing import Any, Dict, Iterable, Optional, cast

from ..algorithms.registry import ApplicationsRegistry
from .algorithms import (
    AVAILABLE_ALGORITHMS,
    AVAILABLE_ALGORITHMS_CATEGORIES,
    filter_algorithm_applications,
)
from .argument_parser import ArgumentParser, DataClassType

logger = logging.getLogger(__name__)


@dataclass
class AlgorithmApplicationArguments:
    """Algorithm application arguments."""

    __name__ = "algorithm_base_args"

    algorithm_type: Optional[str] = field(
        default=None,
        metadata={
            "help": f"Inference algorithm type, supported types: {', '.join(AVAILABLE_ALGORITHMS_CATEGORIES['algorithm_type'])}."
        },
    )
    domain: Optional[str] = field(
        default=None,
        metadata={
            "help": f"Domain of the inference algorithm, supported types: {', '.join(AVAILABLE_ALGORITHMS_CATEGORIES['domain'])}."
        },
    )
    algorithm_name: Optional[str] = field(
        default=None,
        metadata={"help": "Inference algorithm name."},
    )
    algorithm_application: Optional[str] = field(
        default=None,
        metadata={"help": "Inference algorithm application."},
    )
    algorithm_version: Optional[str] = field(
        default=None,
        metadata={"help": "Inference algorithm version."},
    )


@dataclass
class InferenceArguments:
    """Inference arguments."""

    __name__ = "inference_base_args"

    target: Optional[str] = field(
        default=None,
        metadata={
            "help": (
                "Optional target for generation represented as a string. Defaults to None, it can be "
                "also provided in the configuration_file as an object, but the commandline takes precendence."
            )
        },
    )
    number_of_samples: int = field(
        default=5,
        metadata={"help": "Number of generated samples, defaults to 5."},
    )
    configuration_file: Optional[str] = field(
        default=None,
        metadata={
            "help": "Configuration file for the inference pipeline in JSON format."
        },
    )
    print_info: bool = field(
        default=False,
        metadata={
            "help": "Print info for the selected algorithm, preventing inference run. Defaults to False."
        },
    )


def main() -> None:
    """Run an inference pipeline."""
    logging.basicConfig(stream=sys.stdout, level=logging.INFO)

    parser = ArgumentParser(
        cast(
            Iterable[DataClassType], (AlgorithmApplicationArguments, InferenceArguments)
        )
    )
    algorithm_args, inference_args, _ = parser.parse_args_into_dataclasses(
        return_remaining_strings=True
    )

    filters = algorithm_args.__dict__
    matching_algorithms = filter_algorithm_applications(
        algorithms=AVAILABLE_ALGORITHMS, filters=filters
    )
    if len(matching_algorithms) > 1:
        logger.info(
            f"Multiple algorithms matching the parameters:{os.linesep}"
            f"{os.linesep.join(map(str, matching_algorithms))}{os.linesep}"
            f"Select one by specifying additional algorithms parameters: {','.join('--' + key for key, value in filters.items() if not value)}.",
        )
        return
    elif len(matching_algorithms) < 1:
        provided_filters = {key: value for key, value in filters.items() if value}
        logger.error(
            "No algorithms matching the provided parameters, "
            f"please review the supported algorithms:{os.linesep}"
            f"{os.linesep.join(map(str, matching_algorithms))}{os.linesep}"
            f"Please review the parameters provided:{os.linesep}"
            f"{provided_filters}"
        )
    selected_algorithm = matching_algorithms[0]
    logger.info(f"Selected algorithm: {selected_algorithm}")
    target = inference_args.target
    number_of_samples = inference_args.number_of_samples
    print_info = inference_args.print_info
    configuration_filepath = inference_args.configuration_file

    if print_info:
        algorithm_configuration = ApplicationsRegistry.get_configuration_instance(
            **selected_algorithm
        )
        algorithm_configuration_dict = {**algorithm_configuration.to_dict()}
        _ = algorithm_configuration_dict.pop("description", None)
        logger.info(
            f"Selected algorithm support the following configuration parameters:{os.linesep}"
            f"{json.dumps(algorithm_configuration_dict, indent=1)}{os.linesep}"
            f"Target information:{os.linesep}"
            f"{json.dumps({'target': algorithm_configuration.get_target_description()}, indent=1)}"
        )
        return

    configuration: Dict[str, Any] = {}

    if configuration_filepath is not None:
        with open(configuration_filepath) as fp:
            configuration = json.load(fp)
    else:
        logger.info("No configuration file provided, running using default parameters.")

    if target is not None:
        if "target" in configuration:
            logger.info(
                "Target provided both via commandline and configuration file. "
                f"The commandline one will be used: {target}."
            )
        configuration["target"] = target

    algorithm = ApplicationsRegistry.get_application_instance(
        target=configuration.pop("target", None),
        **selected_algorithm,
        **configuration,
    )
    logger.info(
        f"Starting generation with the following configuration:{algorithm.configuration}"
    )
    print(
        f"{os.linesep.join(map(str, algorithm.sample(number_of_items=number_of_samples)))}"
    )


if __name__ == "__main__":
    main()
